"""
A very convenient library, which contains:

    TonyDBC: A context manager class for MariaDB

To protect certain databases against accidental deletion, please set in your .env:

    PRODUCTION_DATABASES = ["important_db", "other_important_db"]

To instrument TonyDBC to diagnose performance:

     simply populate the following in .env:

        AUDIT_PATH = C:\\tony_log.csv

    or to populate ONLY the database table `tony`:

        AUDIT_PATH = database

    or to disable performance logging:

        AUDIT_PATH =

DOCUMENTATION

The common core methods of TonyDBC:

    append_to_table: INSERT; returns DataFrame (optionally)
        table, df, return_reindexed=False
    query_table:     SELECT; returns a DataFrame
        table, query=None
    execute:         does not return any data
        command, command_values=None, before_retry_cmd=None
    get_data:        SELECT; returns a list of dicts.
        query
    update_blob:     UPDATE
        table_name, blob_column, id_value, filepath, max_size_MB=16

"""

import code
import copy
import inspect
import os
import pathlib
import pickle
import queue
import random
import shutil
import sys
import threading
import time
import zoneinfo
from contextlib import contextmanager
from types import TracebackType
from typing import Any, Callable, Literal, overload

import dateutil
import filelock

# Use the official mariadb Python connection library
import mariadb  # type: ignore
import pandas as pd
import tzlocal
import uuid6
from mariadb.constants.CLIENT import MULTI_STATEMENTS  # type: ignore

from .dataframe_fast import DATATYPE_MAP, FIELD_TYPE_DICT, DataFrameFast, read_sql_table
from .env_utils import get_env_list
from .tony_utils import (
    deserialize_table,
    get_next_word_after_from,
    get_payload_info,
    get_tz_offset,
    prepare_scripts,
    serialize_table,
)

# Maximum characters to output to `tony`.`query`
QUERY_AUDIT_MAX_CHARS = 1000

# Default to these environment variables if no credentials are provided
DEFAULT_CREDENTIALS = {
    "host": "MYSQL_HOST",
    "user": "MYSQL_READWRITE_USER",
    "password": "MYSQL_READWRITE_PASSWORD",
    "database": "MYSQL_DATABASE",
}

# Max number of times to re-try a command if connection is lost
MAX_RECONNECTION_ATTEMPTS = 3


def get_currentframe_method() -> str:
    cur_frame = inspect.currentframe()
    if cur_frame and hasattr(cur_frame, "f_code"):
        return cur_frame.f_code.co_name
    else:
        return "unknown"


def check_connection(fn: Callable) -> Callable:
    def conn_wrapper(self: "_TonyDBCOnlineOnly", *args: Any, **kwargs: Any) -> Any:
        try:
            self._mariatonydbcn.ping()
        except mariadb.Error:
            _ping_str = "Ping failed: Restarting mariadb connection"
            self.log(_ping_str)
            self.__enter__()
        result = fn(self, *args, **kwargs)
        return result

    return conn_wrapper


class _TonyDBCOnlineOnly:
    """
    Generic context manager for a database connection and for several
    common queries and commands made against a database.

    Creates a persistent database connection, using the mariadb connector,
        the official Python connection library for mariadb

    You can change the database it points to by running .use(new_database)

    """

    # Type annotations for instance attributes
    host: str
    user: str
    password: str
    database: str
    port: int
    session_uuid: str
    session_timezone: str
    ipath: str
    media_to_deserialize: dict[str, list[str]]
    _audit_db: "_TonyDBCOnlineOnly | None"

    def __init__(
        self,
        host: str | None = None,
        user: str | None = None,
        password: str | None = None,
        database: str | None = None,
        port: int = 3306,
        media_to_deserialize: dict[str, list[str]] = {},
        autocommit: bool = True,
        logger_instance: Any | None = None,
        prefix: str = "",
        lost_connection_callback: Callable | None = None,
        session_timezone: str | None = None,
        interact_after_error: bool = False,
        force_no_audit: bool = False,
    ) -> None:
        """
        Parameters:
            media_to_deserialize: dict of tables with lists of columns to deserialize  e.g.
                {
                    "location_image": [
                        "bounds_i",
                        "bounds_w",
                        "bounds_clipped_w",
                        "big_box_bounds_clipped_i",
                    ]
                }

            lost_internet_callback: this function will be called if the
                                    connection to the database is lost
                                    while running an SQL command.
        """
        for field, env_key in DEFAULT_CREDENTIALS.items():
            param_value = locals().get(field, None)
            if param_value is None:
                # Try to grab the default
                if env_key in os.environ:
                    setattr(self, field, os.environ[env_key])
                else:
                    raise AssertionError(
                        f"TonyDBC: Not all credentials provided: "
                        f" e.g. {field} not provided and not in os.environ.  You must provide {DEFAULT_CREDENTIALS}"
                    )
            else:
                setattr(self, field, param_value)

            val = getattr(self, field, "")
            assert isinstance(val, str), (
                f"TonyDBC: Credential {field} ({val}) should be a string, not {type(val).__name__}"
            )
            assert len(val) > 0, (
                f"TonyDBC: Credential {field} ({val}) should be a string of length > 0."
            )

        # uuid will be set when the connection is made
        self.session_uuid = str(uuid6.uuid8())
        self.port = port
        self.media_to_deserialize = media_to_deserialize
        self.autocommit = autocommit
        self._l = logger_instance
        self._lost_connection_callback = lost_connection_callback
        self.using_temp_conn = False
        self.interact_after_error = interact_after_error

        # Used to preface all logging statements
        self.prefix = prefix

        self.prelim_session_timezone = session_timezone

        self.do_audit = False
        self._audit_db = None  # Separate TonyDBC instance for audit operations

        # For debugging purposes, we may wish to instrument all queries
        if (not force_no_audit) and "AUDIT_PATH" in os.environ:
            ipath = os.environ["AUDIT_PATH"]
            if ipath == "database":
                # Track only on the database, not locally
                self.do_audit = True
                self.ipath = ipath
            elif ipath != "":
                self.do_audit = True
                self.ipath = str(pathlib.Path(ipath).resolve())
                # Delete the instrumentation csv file if it has zero size
                if (
                    pathlib.Path(self.ipath).exists()
                    and pathlib.Path(self.ipath).stat().st_size == 0
                ):
                    pathlib.Path(self.ipath).unlink()

    def __enter__(self) -> "_TonyDBCOnlineOnly":
        self.make_connection()

        if self.do_audit:
            # Create a separate database connection specifically for audit operations
            # (we do this to avoid interfering with last_insert_id in the main connection)
            self.log(
                "Creating separate audit connection to preserve last_insert_id integrity."
            )
            # Create a separate _TonyDBCOnlineOnly instance for audit operations
            self._audit_db = _TonyDBCOnlineOnly(
                host=self.host,
                user=self.user,
                password=self.password,
                database=self.database,
                autocommit=True,  # Always use autocommit for audit operations
                lost_connection_callback=self._lost_connection_callback,
                session_timezone=self.session_timezone,
                interact_after_error=self.interact_after_error,
                # THIS IS CRUCIAL TO AVOID RECURSIVELY SETTING UP INFINITE AUDIT CONNECTIONS:
                force_no_audit=True,
            )
            # Initialize the audit connection
            self._audit_db.__enter__()

        return self

    def make_connection(self) -> None:
        self.log(f"Connecting to database {self.database} on {self.host}.")
        num_attempts = MAX_RECONNECTION_ATTEMPTS
        while True:
            if num_attempts < MAX_RECONNECTION_ATTEMPTS:
                self.log(
                    f"TonyDBC.__enter__ {num_attempts} / {MAX_RECONNECTION_ATTEMPTS} remaining"
                )
            try:
                # DOCS:
                # https://github.com/mariadb-corporation/mariadb-connector-python/blob/f26934540d9506b6079ad92f603b697c761622de/mariadb/mariatonydbcnection.c#L301
                self._mariatonydbcn = mariadb.connect(
                    host=self.host,
                    port=self.port,
                    user=self.user,
                    password=self.password,
                    database=self.database,
                    client_flag=MULTI_STATEMENTS,
                    autocommit=self.autocommit,
                    # connect_timeout=30,  # Default is 0, in seconds (have not yet tried this)
                    read_timeout=3600,  # 15,  # in seconds
                    write_timeout=3600,  # 20  # in seconds
                    local_infile=True,
                    compress=True,
                )
            except mariadb.InterfaceError as e:
                if num_attempts > 0:
                    num_attempts -= 1
                    self.log(f"num_attempts {num_attempts} here")
                    continue

                raise ConnectionError(
                    f"Could not connect to database. {e}\n"
                    f"Possible causes: \n"
                    f"(1) Your local Internet connection is down\n"
                    f"(2) The server {self.host} is down\n"
                    f"(3) The mariadb docker container on the server is down\n"
                    f"(4) You are using a VPN.  Disable VPN please."
                )
            except mariadb.OperationalError as e:
                # e.g. mariadb.OperationalError: Can't connect to server on 'fling.ninja' (10060)
                self.log(f"Got mariadb.OperationalError {e}")

                if num_attempts > 0:
                    num_attempts -= 1
                    self.log(f"num_attempts {num_attempts} here")
                    continue

                if self._lost_connection_callback is not None:
                    self.log(
                        "mariadb.OperationalError during initial connect(): "
                        "Calling the callback provided for this situation."
                    )
                    self._lost_connection_callback()

                raise ConnectionError(
                    f"Database server is down?  Got mariadb.OperationalError {e}"
                )
            break

        # If the connection fails, attempt to reconnect (cool!)
        # but sadly due to issue https://github.com/Fling-Asia/tonydbc/issues/7
        # this will reset local_infile = False so we cannot use it.
        # instead we rely on our lost_connection_callback to reconnect
        self._mariatonydbcn.auto_reconnect = False

        if not self._mariatonydbcn.open:
            if self._lost_connection_callback is not None:
                self.log("not self._mariatonydbcn.open so calling the callback")
                self._lost_connection_callback()

            raise AssertionError("Database could not open via mariadb")

        # Set the timezone for the first time
        self.set_timezone(self.prelim_session_timezone)

        self.session_uuid = str(uuid6.uuid8())

        if self.do_audit:
            # Create the `tony` table if it doesn't exist
            current_dir = os.path.dirname(os.path.abspath(__file__))

            # Construct the full path to the log.txt file
            sql_path = pathlib.Path(current_dir) / "audit_table.sql"
            assert sql_path.is_file()
            cmd = prepare_scripts(self.database, [str(sql_path)])

            self.execute(cmd, no_tracking=True)

    def set_timezone(self, session_timezone: str | None = None) -> "_TonyDBCOnlineOnly":
        """
        Set the IANA session time zone for display and input purposes for TIMEZONE fields
        e.g. session_timezone = "Asia/Bangkok" => session_time_offset = "+07:00"

        if None specified, default to os.environ["DEFAULT_TIMEZONE"]
        """
        if session_timezone is None:
            session_timezone = os.environ["DEFAULT_TIMEZONE"]
            self.log(f"set_timezone defaulting to DEFAULT_TIMEZONE {session_timezone}")
            assert session_timezone is not None, "Error: DEFAULT_TIMEZONE is None"

        if not hasattr(self, "session_timezone"):
            self.session_timezone = session_timezone
            self.log(
                f"session_timezone was not set; now set to {self.session_timezone}"
            )
        elif session_timezone == self.session_timezone:
            self.log(f"session_timezone unchanged at {self.session_timezone}")
            return self
        else:
            self.log(
                f"session_timezone changing from {self.session_timezone} -> {session_timezone}"
            )
            self.session_timezone = session_timezone

        if self.session_timezone not in zoneinfo.available_timezones():
            raise AssertionError(
                f"The session_timezone specified, {self.session_timezone}, "
                "is not a valid IANA time zone (for a complete list,"
                " see zoneinfo.available_timezones())."
            )

        system_timezone = str(tzlocal.get_localzone())

        if self.session_timezone != system_timezone:
            self.log(
                f"WARNING: The session timezone specified, {self.session_timezone}, "
                f"is not the same as the system time "
                f"zone {system_timezone}.  This might be fine if you are rendering data for another time zone, "
                "but it WILL cause problems if you are APPENDING or INSERTING data if you are not careful."
            )

        self.session_time_offset = get_tz_offset(
            iana_tz=self.session_timezone
        )  # e.g. "+07:00"

        # Set the database session to be this time zone
        # any INSERT or UPDATE commands executed on TIMEZONE data types will assume
        # the data was expressed in this time zone.
        self.execute(
            f"SET @@session.time_zone:='{self.session_time_offset}';", no_tracking=True
        )
        self.default_tz = dateutil.tz.gettz(self.session_timezone)
        assert self.default_tz is not None

        self.log(
            f"TonyDBC: @@session.time_zone set to {self.session_time_offset}, "
            f"which is IANA timezone {self.session_timezone}."
        )

        # e.g. '+00:00'
        timezone_data = self.get_data("SELECT @@session.time_zone;")
        actual_time_offset = timezone_data[0]["@@session.time_zone"]
        assert actual_time_offset == self.session_time_offset, (
            f"Actual timezone offset {actual_time_offset} does not match expected timezone offset {self.session_time_offset} of timezone {self.session_timezone}"
        )

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        # Commit all pending transactions if necessary
        if not self.autocommit:
            self.log("TonyDBC commit pending transactions.")
            self.commit()
        # Only close if connection is still open
        if (
            hasattr(self, "_mariatonydbcn")
            and self._mariatonydbcn
            and not getattr(self._mariatonydbcn, "_closed", True)
        ):
            self._mariatonydbcn.close()
            self.log("TonyDBC mariadb connection closed.")
        else:
            self.log("TonyDBC mariadb connection already closed.")

        if self.do_audit:
            assert self._audit_db is not None, (
                "Audit connection is not available at __exit__ to shut down"
            )

            # Close the separate audit connection
            self._audit_db.__exit__(None, None, None)
            self.log("TonyDBC audit connection closed.")
            self._audit_db = None

            self.log(f"TonyDBC debug logs at: `{self.host}`.`{self.database}`.`tony`")
            if str(self.ipath) != "database":
                self.log(f"TonyDBC debug logs also at: {self.ipath}")

        if isinstance(exc_type, type(SystemExit)):
            self.log("TonyDBC: user typed exit() in interpreter.")
        elif exc_type is not None:
            self.log(
                f"TonyDBC: exception triggered __exit__: \n"
                f"exc_type: {exc_type}\nvalue: {exc_val}\ntraceback: {exc_tb}"
            )
            return None  # Do not handle the exception; propagate it up
        else:
            self.log("TonyDBC: normal __exit__ successful.")
            # (No need to return a value since `exit_type` is None.)

    def start_temp_conn(self) -> None:
        """Start using a temporary database connection"""
        if self.using_temp_conn:
            raise AssertionError("You are already using your temporary connection.")
        self.using_temp_conn = True
        self._mariatonydbcn_old = self._mariatonydbcn
        # Create a new connection
        self.__enter__()

    def close_temp_conn(self) -> None:
        """Stop using the temporary database connection"""
        if not self.using_temp_conn:
            raise AssertionError("You are not using your temporary connection.")
        # Close the temporary connection
        self.__exit__(None, None, None)
        # Restore the original connection
        self._mariatonydbcn = self._mariatonydbcn_old
        self.using_temp_conn = False

    def now(self) -> pd.Timestamp:
        return pd.Timestamp.now(tz=self.session_timezone)

    def cursor(self) -> mariadb.Cursor:
        return self._mariatonydbcn.cursor()

    def begin_transaction(self) -> None:
        """This must be followed by a commit later.  Turns off autocommit."""
        # We must ensure autocommit is not on to be able to create a transaction
        if self.autocommit:
            self.execute("SET AUTOCOMMIT=0;")

        self._mariatonydbcn.begin()

    def commit(self) -> None:
        try:
            self._mariatonydbcn.commit()
        except mariadb.InterfaceError as e:
            # e.g. Lost connection to server during query
            if self._lost_connection_callback is not None:
                self.log(
                    "mariadb.InterfaceError during commit: "
                    "Calling the callback provided for this situation."
                )
                self._lost_connection_callback()

            raise ConnectionError(
                f"Database server is down?  Got mariadb.InterfaceError {e}"
            )

        # Turn autocommit back on if that's what was originally requested
        if self.autocommit:
            self.execute("SET AUTOCOMMIT=1;")

    def log(self, msg: str):
        """Internal logging; use normal print if a logger was not provided"""
        if self.prefix != "":
            msg = f"{self.prefix} | {msg}"

        if self._l is None:
            print(msg)
        else:
            self._l.info(msg)

    def column_info(self, table_name: str | None = None) -> pd.DataFrame:
        """Returns the column information.
        Parameters:
            table_name: string.  If None, returns column info for ALL tables.
        """
        clauses = [f"TABLE_SCHEMA = '{self.database}'"]
        if table_name is not None:
            clauses.append(f"TABLE_NAME = '{table_name}'")

        r = self.get_data(
            f"""
            SELECT *
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE {"AND ".join(clauses)};
            """,
            no_tracking=True,
        )

        return pd.DataFrame(r)

    @property
    def last_insert_id(self) -> int:
        """Get the last insert ID from the main connection.

        Note: This is now stable even when audit is enabled, because audit operations
        use a separate database connection that doesn't interfere with the main connection's
        last_insert_id value.
        """
        data = self.get_data("SELECT LAST_INSERT_ID() AS id;", no_tracking=True)
        return data[0]["id"]

    def use(self, new_database: str) -> None:
        """Change databases"""
        self.database = new_database
        # Reset current connections
        self.__exit__(None, None, None)
        # Make new connections
        self.__enter__()

    def iso_timestamp_to_session_time_zone(self, iso_timestamp_string: str):
        """e.g. converts an ISO-formatted string to the database session's time zone,
        self.default_tz, which is useful when adding to a mariadb TIMESTAMP
        which assumes the time is formatted as the session time zone.
        e.g. self.default_tz = "Asia/Bangkok",
            "2023-03-03T09:00:00+07:00" ->
            "2023-03-03 09:00:00"
                (since 9-7+7 = 9)

            or

            "2023-03-03T09:00:00+03:00" ->
            "2023-03-03 13:00:00"
                (since 9-3+7 = 13)
        """
        try:
            ts = dateutil.parser.parse(iso_timestamp_string)
        except dateutil.parser._parser.ParserError as e:
            raise ValueError(f"Timestamp '{iso_timestamp_string}' is invalid. {e}")
        else:
            iso_ts = ts.astimezone(self.default_tz)  # dateutil.tz.UTC
            return iso_ts.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]

    def write_dataframe(self, df, table_name, if_exists="replace", index=False):
        """Write all rows of a dataframe to the database
        Assumes the fields are the same in number and data type
        """
        df0 = DataFrameFast(df)
        df0.to_sql_fast(
            name=table_name,
            con=self._mariatonydbcn,
            session_timezone=self.session_timezone,
            if_exists=if_exists,
            index=index,
        )

    def update_table(self, table_name: str, df):
        """Perform an UPDATE with appropriate serialization.

        This is the spiritual sister of `append_to_table` (i.e. it uses the same fancy semantics
        to serialize before inserting, which is useful for JSON-ifying columns like the `OCR` column)

        e.g. if one of the fields is OCR, it will serialize as a string: "['hello', 'world']"

        Parameters:
            table_name is the name of the table
            df: is a dataframe containing the correct indexes to update, and
                containing the correct columns to be updated

        """
        if len(df) == 0:
            self.log(f"UPDATE 0 rows in {self.database}.{table_name}")
            return

        pk = self.get_primary_key(table=table_name, default="id")
        # For vanity's sake, give the dataframe index the correct name
        df.index.name = pk
        # We don't try to cast the primary key (which is always an int)
        col_dtypes = {
            k: v
            for k, v in self.get_column_datatypes(table=table_name).items()
            if k != pk
        }

        if table_name in self.media_to_deserialize:
            columns_to_serialize = self.media_to_deserialize[table_name]
        else:
            columns_to_serialize = []

        df_serialized = serialize_table(df, col_dtypes, columns_to_serialize)

        # Append our values to the actual database table
        self.log(
            f"UPDATE {len(df_serialized)} rows in {self.database}.{table_name} "
            f"for columns: {df_serialized.columns}"
        )

        # e.g. 'UPDATE roi SET OCR = ? WHERE `id` = ?'
        update_command = f"""
            UPDATE {table_name}
            SET {", ".join([v + " = ?" for v in df_serialized.columns])}
            WHERE `id` = ?
            """

        # Make a list of tuples of all the values to be changed.
        # Put the index LAST in each tuple
        update_command_values = [
            tuple(row) + (i,) for i, row in df_serialized.iterrows()
        ]

        with self.cursor() as cursor:
            cursor.executemany(update_command, update_command_values)

    def update_blob(self, table_name, blob_column, id_value, filepath, max_size_MB=16):
        """Add a BLOB (Binary Large OBject) to the database, from a file
        This requires that the table row already exists; we are just updating it

        Note, to check for max size, you have to know the actual blob type and then specify
        the size manually since we don't actually check the blob type.  TODO: check automatically

        For now, we assume MEDIUMBLOB (16 MB)

        max size for a BLOB is 64 KB
        max size for a MEDIUMBLOB is 16 MB
        max size for a LONGBLOB is 4 GB
        """
        assert os.path.isfile(filepath)

        # Check the file size of the blob
        file_size = os.path.getsize(filepath)

        # Convert size to MB
        file_size_MB = file_size // (1024**2)

        if file_size_MB > max_size_MB:
            raise AssertionError(
                f"The file {filepath} to be uploaded as a blob to {table_name}.{blob_column} is too large. "
                f" It is {file_size_MB} MB, but the maximum allowed size is {max_size_MB} MB."
            )

        # Open your file in binary mode
        with open(filepath, "rb") as blob_file:
            blob = blob_file.read()

        # Update the relevant row, adding the BLOB
        blob_mdb_obj = mariadb.Binary(blob)
        self.execute(
            command=f"UPDATE {table_name} SET {blob_column} = ? WHERE `id` = {id_value}",
            command_values=(blob_mdb_obj,),
        )

    def read_dataframe_from_table(
        self, table_name: str, query: str, columns_to_deserialize: list[str] = []
    ) -> pd.DataFrame:
        """Read the dataframe but don't set the index"""
        full_name = ".".join([self.database, table_name])
        df = read_sql_table(name=full_name, con=self, query=query)

        primary_key = self.get_primary_key(table=table_name, default="id")

        # If the primary key was selected, then use it, otherwise leave the table with no index
        if primary_key in df.columns:
            df.set_index(primary_key, inplace=True)
            df.sort_index(inplace=True)

        # DESERIALIZE!
        # (I wrote this but it's not needed; only arrays need special deserialization
        # and this cannot be inferred from the SQL datatype which will be varchar)  :(
        # Query INFORMATION_SCHEMA to get column data types
        # col_df = self.column_info(con=self._mariatonydbcn)
        # nonstring_filter = (col_df['DATA_TYPE'] != 'varchar')
        # Deserialize all columns in tables which are NOT varchar
        # table_filter = (col_df['TABLE_NAME'] == table)
        # cols = list(col_df.loc[
        #    table_filter & nonstring_filter,
        #    'COLUMN_NAME'])
        df = deserialize_table(
            df,
            columns_to_deserialize=columns_to_deserialize,
            session_timezone=self.session_timezone,
        )

        return df

    @overload
    def get_data(
        self,
        query: str,
        before_retry_cmd: str | None = ...,
        no_tracking: bool = ...,
        return_type_codes: Literal[False] = ...,
    ) -> list[dict[str, Any]]: ...

    @overload
    def get_data(
        self,
        query: str,
        before_retry_cmd: str | None = ...,
        no_tracking: bool = ...,
        return_type_codes: Literal[True] = ...,
    ) -> tuple[list[dict[str, Any]], dict[Any, str]]: ...

    def get_data(
        self,
        query: str,
        before_retry_cmd: str | None = None,
        no_tracking: bool = False,
        return_type_codes: bool = False,
    ) -> Any:
        if self.do_audit and not no_tracking:
            started_at = self.now()

        attempts_remaining = MAX_RECONNECTION_ATTEMPTS
        while attempts_remaining > 0:
            with self.cursor() as cursor:
                try:
                    # TODO: Consider adding
                    # cursor.execute("SET time_zone = '+00:00'")
                    # every time though???
                    # https://stackoverflow.com/questions/1136437/
                    if (
                        before_retry_cmd is not None
                        and attempts_remaining < MAX_RECONNECTION_ATTEMPTS
                    ):
                        cursor.execute(before_retry_cmd)

                    cursor.execute(query)
                    records0 = cursor.fetchall()
                    fields = cursor.description
                    """
                    Cursor.description is an 11-item tuple with the following:
                        name
                        type_code   can be looked up via FIELD_TYPE_DICT
                        display_size
                        internal_size
                        precision
                        scale
                        null_ok
                        field_flags
                        table_name
                        original_column_name
                        original_table_name
                    """
                except mariadb.InterfaceError:
                    self.log(
                        f"Reconnecting to mariadb; attempting query again. "
                        f"Attempts remaining {attempts_remaining} BEFORE this attempt."
                    )
                    self.make_connection()
                    attempts_remaining -= 1
                except Exception as e:
                    if self.interact_after_error:
                        self.log(
                            f"mariadb execute command failed: {query} with error {e}"
                        )
                        code.interact(local=locals(), banner=f"{e}")
                    else:
                        raise Exception(
                            f"mariadb execute command failed: {query} with error {e}"
                        )
                else:
                    # if False and cursor.lastrowid is None:
                    #    raise AssertionError(
                    #        f"An error occurred with one of the commands {command}; lastrowid is None"
                    #    )
                    break
                try:
                    cursor.execute("SHOW WARNINGS;")
                except mariadb.InterfaceError as e:
                    self.log(
                        f"Cannot run cursor.execute('SHOW WARNINGS;') because mariadb.InterfaceError {e}."
                    )
                    warnings = []
                else:
                    # Check for warnings
                    warnings = cursor.fetchall()

            if warnings:
                for warning in warnings:
                    self.log(warning)

        records = [
            {fields[i][0]: field_value for i, field_value in enumerate(v)}
            for v in records0
        ]

        if self.do_audit and not no_tracking:
            self._save_instrumentation(
                method=get_currentframe_method(),
                table=None,
                query=query,
                started_at=started_at,
                **get_payload_info(records),
            )

        if return_type_codes:
            # Get a list of dicts with proper field names
            # (i.e. records in the pandas sense)
            # (e.g. {'media_object_id': 'LONGLONG', 'full_path': 'VAR_STRING'}
            type_codes = {v[0]: FIELD_TYPE_DICT[v[1]] for v in fields}

            return records, type_codes
        else:
            return records

    @property
    def databases(self):
        return [d["Database"] for d in self.get_data("SHOW DATABASES;")]

    @property
    def users(self):
        return [
            (d["Host"], d["User"])
            for d in self.get_data("SELECT * FROM mysql.user", no_tracking=True)
        ]

    def show_grants(self, username: str, host="%"):
        return self.get_data(
            f"SHOW GRANTS FOR '{username}'@'{host}';", no_tracking=True
        )

    @property
    def production_databases(self):
        try:
            return self.__production_databases
        except AttributeError:
            # Protect these databases - test harness will check to ensure
            # it doesn't try to drop these databases (will raise AssertionError)
            if "PRODUCTION_DATABASES" in os.environ:
                dbs = get_env_list("PRODUCTION_DATABASES")
            else:
                dbs = []

            dbs += [
                "information_schema",
                "mysql",
                "performance_schema",
                "sys",
            ]
            # Remove duplicates
            self.__production_databases = sorted(set(dbs))

            return self.__production_databases

    def drop_database(self, database):
        if database in self.production_databases:
            raise AssertionError(
                f"DANGER DANGER!  You are trying to drop {database}, "
                "a production database. Please talk to your manager for advice on what "
                " to do.  Do NOT bypass this error."
            )

        if database in self.databases:
            self.log(f"DROP {database} since it already exists")
            # First unlock the tables so they won't prevent us from dropping the database
            self.execute(f"USE {database}; UNLOCK TABLES;")
            self.execute(f"DROP DATABASE {database};")

    def post_data(self, query: str):
        with self.cursor() as cursor:
            record = cursor.execute(query)

        return record

    def post_datalist(self, query: str, insert_data: list[Any]) -> Any:
        with self.cursor() as cursor:
            record = cursor.executemany(query, insert_data)
        return record

    def execute(
        self,
        command: str,
        command_values: tuple | None = None,
        before_retry_cmd: str | None = None,
        no_tracking: bool = False,
        log_progress: bool = False,
    ) -> None:
        """Parameters:
        command_values: a tuple of values [optional]
        """
        if log_progress:
            self.log(f"Executing {command[:50]}")

            stop_wait_message = (
                threading.Event()
            )  # Event to signal the waiting thread to stop

            def wait_message():
                log_increment = 0
                while not stop_wait_message.is_set():
                    if log_increment >= 50:
                        # Every 50 x 0.1s = 5s, we will notify that the query is still going...
                        self.log(f"Waiting for {command[:50]} ...")
                        log_increment = 0
                    # Sleep for only a short time to avoid slowing things down when we .join
                    # if the query ends early
                    time.sleep(0.1)
                    log_increment += 1

            wait_thread = threading.Thread(target=wait_message, daemon=True)
            wait_thread.start()

        if self.do_audit and not no_tracking:
            started_at = self.now()

        attempts_remaining = MAX_RECONNECTION_ATTEMPTS
        while attempts_remaining > 0:
            with self.cursor() as cursor:
                try:
                    # TODO: Consider adding
                    # cursor.execute("SET time_zone = '+00:00'")
                    # every time though???
                    # https://stackoverflow.com/questions/1136437/
                    if (
                        before_retry_cmd is not None
                        and attempts_remaining < MAX_RECONNECTION_ATTEMPTS
                    ):
                        cursor.execute(before_retry_cmd)
                    if command_values is not None:
                        cursor.execute(command, command_values)
                    else:
                        cursor.execute(command)
                except mariadb.InterfaceError:
                    # Try reconnecting by explicitly calling the TonyDBC enter method
                    # (not any child class overridden version) by doing this instead
                    # of saying self.__enter__()
                    self.log(
                        f"Reconnecting to mariadb; attempting command {command} again. "
                        f"Attempts remaining {attempts_remaining} BEFORE this attempt."
                    )
                    self.make_connection()
                    attempts_remaining -= 1
                except Exception as e:
                    if self.interact_after_error:
                        self.log(
                            f"mariadb execute command failed: {command} with error {e}"
                        )
                        code.interact(local=locals(), banner=f"{e}")
                    else:
                        if log_progress:
                            stop_wait_message.set()
                            wait_thread.join()
                        raise Exception(e)
                else:
                    # if False and cursor.lastrowid is None:
                    #    raise AssertionError(
                    #        f"An error occurred with one of the commands {command}; lastrowid is None"
                    #    )
                    break
                cursor.execute("SHOW WARNINGS;")
                # Check for warnings
                warnings = cursor.fetchall()

            if warnings:
                for warning in warnings:
                    self.log(warning)

            # mariadb.InterfaceError: Lost connection to server during query
            # mariadb.OperationalError: Can't connect to server on 'fling.ninja' (10060)
            # mariadb.InterfaceError: Server has gone away

        # Signal the waiting thread to stop
        if log_progress:
            stop_wait_message.set()
            wait_thread.join()
            self.log(f"Executing {command[:50]} - DONE")

        if self.do_audit and not no_tracking:
            self._save_instrumentation(
                method=get_currentframe_method(),
                table=None,
                query=command,
                started_at=started_at,
                **get_payload_info(command_values),
            )

    def execute_script(
        self, script_path: str, get_return_values=False, cur_database=None
    ):
        return_values: list[list[dict[str, Any]] | None] = []
        # Read the SQL schema file
        with open(script_path, "r") as file:
            script_string = file.read()

        if cur_database is None:
            cur_database = self.database

        # KLUDGE
        if "DELIMITER" in script_string:
            self.log(
                "Sorry, the python Mariadb connector doesn't seem to be "
                "able to handle the DELIMITER directive.  Can you please go "
                "to HeidiSQL or another MySQL terminal and run this script "
                "manually.  Once it's completed, come back here and press Ctrl-Z."
                f"\nFirst run: USE {cur_database};"
                f"\nScript to run: {script_path}"
            )
            # Wait for user command
            code.interact(local=locals(), banner="Run script manually please")

            if get_return_values:
                return []
            else:
                return

        with self.cursor() as cursor:
            # Get a list of each separate command
            commands = [c + ";" for c in script_string.split(";") if len(c) > 0]
            for command in commands:
                # Ensure command is not simply empty
                command = command.strip()
                if len(command) == 0 or command == ";":
                    continue
                cursor.execute(command)

                if get_return_values:
                    if cursor.description is None:
                        # Handle the case of a command with no result set;
                        # such as `USE fling_db;` in this case fetchall() will
                        # raise a ProgrammingError so we should check for that first
                        return_values.append(None)
                    else:
                        records0 = cursor.fetchall()
                        # Get the field names
                        fields = cursor.description
                        # Get a list of dicts with proper field names
                        # (i.e. records in the pandas sense)
                        records1 = [
                            {
                                fields[i][0]: field_value
                                for i, field_value in enumerate(v)
                            }
                            for v in records0
                        ]
                        return_values.append(records1)

        if get_return_values:
            return return_values

    def get_primary_key(self, table: str, default: str | None = None):
        try:
            return self.primary_keys[table]
        except KeyError as _:
            # Try refreshing before returning the error
            self.refresh_primary_keys()
            try:
                return self.primary_keys[table]
            except KeyError as _:
                # Maybe it's a temporary table or something; try to DESCRIBE it.
                result = self.get_data(f"DESCRIBE {table};", no_tracking=True)
                if len(result) == 0:
                    raise KeyError(f"Table {table} does not seem to exist.")
                primary_keys = [col for col in result if col["Key"] == "PRI"]
                if len(primary_keys) == 0:
                    if default is not None:
                        return default
                    raise KeyError(
                        f"Table {table} is not in our list of tables with "
                        f"primary keys: {self.primary_keys}.  (And no default was provided)"
                    )
                elif len(primary_keys) > 1:
                    raise KeyError(f"Table {table} has more than one primary key!")
                else:
                    return primary_keys[0]["Field"]

    def refresh_primary_keys(self):
        r = self.get_data(
            query=f"""
            SELECT TABLE_NAME, COLUMN_NAME
            FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
            WHERE
                TABLE_SCHEMA = '{self.database}' AND
                CONSTRAINT_NAME = 'PRIMARY'
            """,
            no_tracking=True,
        )
        self._primary_keys = {v["TABLE_NAME"]: v["COLUMN_NAME"] for v in r}

    @property
    def primary_keys(self):
        """Returns a dict of all tables in the database
        and the field which is the primary key  e.g. {'drone': 'id'}
        NOTE: assumes only one primary key
        """
        try:
            return self._primary_keys
        except AttributeError:
            self.refresh_primary_keys()
            return self._primary_keys

    def get_column_datatypes(self, table: str) -> dict[str, Any]:
        try:
            return self.column_datatypes[table]
        except KeyError:
            # Maybe it's a temporary table or something; try to DESCRIBE it.
            result = self.get_data(f"DESCRIBE {table};")
            if len(result) == 0:
                raise KeyError(f"Table {table} does not seem to exist.")
            # We need to map Types like "bigint(20)" to "bigint" to match our DATATYPE_MAP
            col_datatypes = {
                v["Field"]: DATATYPE_MAP[v["Type"].split("(")[0].strip().lower()]
                for v in result
            }
            return col_datatypes

    @property
    def column_datatypes(self):
        """Returns a dict of dicts with Python data types of each column, e.g.
        {
            "warehouse" : {
                ('warehouse_name', str),
                ('warehouse_version', int)
            },
            ...
        }
        """
        try:
            return self._column_datatypes
        except AttributeError:
            r = self.get_data(
                query=f"""
                SELECT TABLE_NAME, COLUMN_NAME, DATA_TYPE
                FROM INFORMATION_SCHEMA.`COLUMNS`
                WHERE
                    TABLE_SCHEMA = '{self.database}'
                """,
                no_tracking=True,
            )
            tables = {v["TABLE_NAME"] for v in r}
            self._column_datatypes = {
                t: {
                    v["COLUMN_NAME"]: DATATYPE_MAP[v["DATA_TYPE"]]
                    for v in r
                    if v["TABLE_NAME"] == t
                }
                for t in tables
            }
            return self._column_datatypes

    def non_primary_keys(self, table):
        """Returns a list of all non-primary keys for a table"""
        pk = self.get_primary_key(table=table)
        cols = self.get_column_datatypes(table=table).keys()
        return [c for c in cols if c != pk]

    @property
    def connection_params(self):
        """Useful for making a new connection.  A dict of all you need.

        e.g.
        db = TonyDBC(...)  # First connection
        params = db.connection_params
        db2 = TonyDBC(**params)  # Second connection
        """
        return {
            "host": self.host,
            "database": self.database,
            "user": self.user,
            "password": self.password,
            "port": self.port,
            "media_to_deserialize": self.media_to_deserialize,
        }

    def insert_row_all_string(self, table, row_dict):
        """Insert a single row into a table, as all string
        Note that the row_dict is required to be all string or stringifyable
        Note that it does not autocommit
        """
        if len(row_dict) == 0:
            return

        stringified_row_dict = {
            k: str(v)
            for k, v in row_dict.items()
            # Skip any columns which are to be NULL and rely on DEFAULT VALUE because it won't accept
            # NULL aka the special mariadb NULL indicator for NULL values
            if (v is not None) and (str(v) != "None")
        }
        # We must sanitize the values to avoid things like single quotes breaking the INSERT
        # so we will pass a list of values to self.execute instead
        command = f"""
            INSERT INTO {table}
                       ({", ".join(stringified_row_dict.keys())})
                VALUES ({", ".join(["%s" for _ in stringified_row_dict])})
            """
        command_values = list(map(str, stringified_row_dict.values()))
        self.execute(command=command, command_values=command_values)

    @contextmanager
    def temp_id_table(self, id_list):
        """Create a temporary table filled with ids for JOINing purposes"""
        # Generate a random temporary table name
        temp_table_name = f"temp_loc_ids_{random.randint(1000, 9999)}"

        try:
            # Create the temporary table
            self.execute(
                f"CREATE TEMPORARY TABLE {temp_table_name} (`id` BIGINT PRIMARY KEY);"
            )

            # Insert the data into the temporary table
            df = pd.DataFrame(id_list, columns=["id"])
            self.append_to_table(temp_table_name, df)

            # Yield the temporary table name to the context
            yield temp_table_name

        finally:
            # Drop the temporary table when the context ends
            self.execute(f"DROP TEMPORARY TABLE IF EXISTS {temp_table_name};")

    def append_to_table(
        self,
        table: str,
        df: pd.DataFrame,
        return_reindexed: bool = False,
        no_tracking: bool = False,
    ) -> pd.DataFrame | None:
        if self.do_audit and not no_tracking:
            started_at = self.now()

        pk = self.get_primary_key(table=table, default="id")
        # For vanity's sake, give the dataframe index the correct name
        # but really we never actually use this index since we let the
        # database autoincrement during the insert
        df.index.name = pk
        # We don't try to cast the primary key (which is always an int)
        col_dtypes = {
            k: v for k, v in self.get_column_datatypes(table=table).items() if k != pk
        }

        columns_to_serialize: list[str]
        if table in self.media_to_deserialize:
            columns_to_serialize = self.media_to_deserialize[table]
        else:
            columns_to_serialize = []

        df_serialized = serialize_table(df, col_dtypes, columns_to_serialize)
        payload_info = get_payload_info(df_serialized)
        payload_size_MB = payload_info["payload_size"] / (1024**2)

        # Append our values to the actual database table
        self.log(
            f"INSERT {payload_info['num_rows']}r x {payload_info['num_cols']}c "
            f"in {self.database}.{table} ({payload_size_MB:.2f} MB)"
        )
        self.write_dataframe(df_serialized, table, if_exists="append", index=False)

        # Return the dataframe with the actual index AUTOINCREMENTED with the correct numbers
        # This is actually GUARANTEED to work, by the way mariadb does consecutive inserts!
        # cool right?
        if return_reindexed:
            # From testing on 2023-11-24, THIS is correct:
            last_insert_id = self.last_insert_id
            # mypy: df.index is Index[Any]; assigning list is valid at runtime but not in types
            df.index = list(range(last_insert_id, last_insert_id + len(df)))  # type: ignore[assignment]
            # NOT this:
            # df.index = list(range(last_insert_id - len(df) + 1, last_insert_id))
            df.index.name = pk

        if self.do_audit and not no_tracking:
            self._save_instrumentation(
                method=get_currentframe_method(),
                table=table,
                query="INSERT",
                started_at=started_at,
                **payload_info,
            )

        return df if return_reindexed else None

    def query_table(self, table: str, query: str) -> pd.DataFrame:
        """Query a single table and deserialize if necessary"""
        columns_to_deserialize: list[str]
        if table in self.media_to_deserialize:
            columns_to_deserialize = self.media_to_deserialize[table]
        else:
            columns_to_deserialize = []

        started_at = self.now()

        cur_df = self.read_dataframe_from_table(
            table_name=table, query=query, columns_to_deserialize=columns_to_deserialize
        )

        if self.do_audit and (query is not None):
            self._save_instrumentation(
                method=get_currentframe_method(),
                table=table,
                query=query,
                started_at=started_at,
                **get_payload_info(cur_df),
            )

        return cur_df

    def refresh_table(self, table) -> None:
        """Load the ENTIRE table which might be highly inefficient
        but for small scales this should be fine.
        """
        cur_df = self.query_table(table, f"""SELECT * FROM {table};""")
        setattr(self, f"{table}_df", cur_df)

    def log_to_db(self, log_dict: dict) -> None:
        """Save log to database
        This module will get the dictionary of log to print put and add
        to the database 'server_log' table. ALL of the keys need past
        to the function with db_credentials. Dictionary format :
            log_message = {
                'log_module' : log_module,
                'log_state' : "STARTED/ERROR/INFO/WARNING/COMPLETED/NOTIFY,
                'log_event' : 'event of the log or action',
                'log_message' : "log message",
                'log_hostname' : hostname
            }
        ps. to get the log_module : log_module = str(os.path.basename(__file__))
            to get the log_hostname : hostname = socket.gethostname()
        """
        log_template = [
            "log_module",
            "log_state",
            "log_event",
            "log_message",
            "log_hostname",
        ]
        if all(k in log_template for k in log_dict):
            query = f"""
                INSERT INTO server_log (`module`, `state`, `log_event`, `message`, `_host`)
                VALUES ('{str(log_dict["log_module"])}','{str(log_dict["log_state"])}','{str(log_dict["log_event"])}','{str(log_dict["log_message"])}','{str(log_dict["log_hostname"])}');
            """
            query = query.replace("'None'", "null")
            self.execute(query)
            self.log(
                f"{log_dict['log_state']} | {log_dict['log_module']} : {log_dict['log_event']} {log_dict['log_message']} {log_dict['log_hostname']}"
            )
        else:
            if "log_module" in log_dict.keys():
                failed_module = str(log_dict["log_module"])
            else:
                failed_module = "unknown"
            self.log(
                f"ERROR | central_log warning : missing log parameters from - {failed_module}"
            )

    def _save_instrumentation(
        self,
        method: str,
        table: str | None,
        query: str,
        started_at: pd.Timestamp,
        payload_size: int,
        num_rows: int,
        num_cols: int,
    ) -> None:
        """Save debugging information about each query that was run"""
        if not self.do_audit:
            return

        assert str(self.ipath) != ""

        # Save in the session timezone
        started_at = started_at.tz_convert(self.session_timezone)
        completed_at = self.now()
        duration = (completed_at - started_at).total_seconds()

        if (duration is not None) and duration > 0:
            MBps = (payload_size / duration) / (1024 * 1024)
        else:
            MBps = None

        # Extract the table name from the query string if necessary
        if table is None or table == "":
            table = get_next_word_after_from(query)

        payload = {
            "table_name": table,
            "query": query[:QUERY_AUDIT_MAX_CHARS],
            "started_at": started_at,
            "completed_at": completed_at,
            "duration_seconds": duration,
            "payload_size_bytes": payload_size,
            "num_rows": num_rows,
            "num_cols": num_cols,
            "method": method,
            "MBps": MBps,
            "session_uuid": self.session_uuid,
            "host": self.host,
            "database_name": self.database,
            "timezone": self.session_timezone,
        }
        df = pd.DataFrame([payload])
        # Remove newlines in query
        df["query"] = df["query"].str.strip().replace("\n", " ")

        # Use separate audit connection to avoid interfering with last_insert_id
        # Use the audit connection's append_to_table method
        assert self._audit_db is not None
        self._audit_db.append_to_table("tony", df, no_tracking=True)

        # If we aren't just tracking exclusively in the database,
        # then write to a file please
        if str(self.ipath) != "database":
            self.log("SAVING LOCK PATH")
            ipath_path = pathlib.Path(self.ipath)
            lock_path = ipath_path.with_suffix(ipath_path.suffix + ".lock")
            with filelock.FileLock(lock_path):
                try:
                    df.to_csv(
                        str(self.ipath),
                        mode="a",
                        header=(not ipath_path.exists()),
                        index=False,
                    )
                except PermissionError as e:
                    self.log(
                        f"WARNING: Instrumentation file is locked for writing: "
                        f"{self.ipath} {e}\n{payload}"
                    )


# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------


class TonyDBC(_TonyDBCOnlineOnly):
    """Public exposed class offering all base functions plus an offline mode

    Note: this is not a true offline mode; it still needs to be in online mode at the beginning, and end.
    But it will make it easy to enqueue many writes, even across multiple sessions.

    Use TonyDBC context manager as normal.  But if you set self.is_online = False, then it will enqueue
    any db updates until you set it back to self.is_online = True.

    If the TonyDBC exits before you set is_online back to True, then the updates will be pickled.
    If you load TonyDBC again, it will load briefly in online mode then flip to offline mode.

    Warning: Like TonyDBC itself, this is intended to be run in a single thread and not to be run in multiple threads

    Usage example:
        with TonyDBC(**connect_kwargs) as db:
            # now you can read from DB
            # ...

            db.is_online = False
            # write a lot of stuff into the queue...
            # ....

            db.is_online = True  # Flush updates into the database

        with TonyDBC(**connect_kwargs) as db:
            # Automatically, TonyDBC will flush all updates not
            # saved to DB from last connection (if db.is_online = True was not done)
            pass

    """

    def __init__(self, *args, **kwargs) -> None:
        mb: str = "MEDIA_BASE_PATH_PRODUCTION"
        pickle_base_path: str
        if mb in os.environ:
            pickle_base_path = os.environ[mb]
        else:
            # Default to the script path if no pickle path was provided
            pickle_base_path = sys.path[0]

        self.__offline_status: str = "online"
        self.__offline_pickle_path: str = os.path.join(
            pickle_base_path, "dbcon_pickle.PICKLE"
        )
        self.__update_queue: queue.Queue = queue.Queue()
        super().__init__(*args, **kwargs)

    @property
    @check_connection
    def is_online(self) -> bool:
        return self.__offline_status != "offline"

    @is_online.setter
    def is_online(self, value: bool) -> None:
        """If the user sets is_online to True, and it was not before, then flush updates."""
        assert isinstance(value, bool)
        if self.__offline_status == "offline" and value:
            self.__offline_status = "flushing"
            self.flush_updates()
            self.__offline_status = "online"
        elif self.__offline_status == "online" and not value:
            self.__offline_status = "offline"

    def flush_updates(self) -> None:
        """Make all the updates to the tables that we have been saving up"""
        # Pickle our updates in case of error
        try:
            if not self.__update_queue.empty():
                self.log("Pickling before flushing to be safe")
                self.pickle_updates()
                self.log(f"Flushing {self.__update_queue.qsize()} database updates")
                while not self.__update_queue.empty():
                    method, kwargs = self.__update_queue.get()
                    getattr(self, method)(**kwargs)
                self.log("Flushing database updates - DONE")
                self.log("Backing up temp pickle")
                shutil.move(
                    str(self.__offline_pickle_path),
                    str(self.__offline_pickle_path) + ".BAK",
                )

        except AttributeError as e:
            if self.interact_after_error:
                code.interact(banner=f"Bad TonyDBC {e}", local=locals())
            else:
                raise AttributeError(e)

    def __enter__(self) -> "TonyDBC":
        super().__enter__()

        if not os.path.isfile(self.__offline_pickle_path):
            return self

        self.is_online = False
        # Try loading unsaved stuff if it exists
        try:
            with open(self.__offline_pickle_path, "rb") as pickle_file:
                update_list = pickle.load(pickle_file)
        except EOFError:
            self.log(
                f"Deleting corrupt pickle file {self.__offline_pickle_path} that is 0 bytes"
            )
            os.remove(self.__offline_pickle_path)
        else:
            self.log(
                f"Finished loading DB updates pickle {self.__offline_pickle_path}."
            )
            # Now get rid of the pickle so we don't use it again
            shutil.move(
                str(self.__offline_pickle_path),
                str(self.__offline_pickle_path) + ".BAK",
            )

        for v in update_list:
            self.__update_queue.put(v)

        # Now flush the pickled updates
        self.is_online = True

        return self

    def pickle_updates(self) -> None:
        # Pickle our queue
        if not self.__update_queue.empty():
            self.log(
                f"Pickling {self.__update_queue.qsize()} updates "
                f"to {self.__offline_pickle_path}"
            )
            backup_queue: queue.Queue = queue.Queue()
            queue_list: list[Any] = []
            while not self.__update_queue.empty():
                v = self.__update_queue.get()
                queue_list.append(copy.deepcopy(v))
                backup_queue.put(copy.deepcopy(v))

            with open(self.__offline_pickle_path, "wb") as pickle_file:
                pickle.dump(queue_list, pickle_file, protocol=pickle.HIGHEST_PROTOCOL)

            self.__update_queue = backup_queue

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.pickle_updates()

        # Clear the queue since we have now archived it
        self.__update_queue = queue.Queue()

        assert self.__update_queue.empty(), (
            "update_queue is not empty even though we just cleared it"
        )

        # Now that we have cleared the queue, we can go to online mode.
        self.is_online = True

        super().__exit__(exc_type, exc_val, exc_tb)
        return None

    def start_temp_conn(self) -> None:
        if not self.is_online:
            raise AssertionError("start_temp_conn is not supported when not online")
        else:
            super(TonyDBC, self).start_temp_conn()

    def close_temp_conn(self) -> None:
        if not self.is_online:
            raise AssertionError("close_temp_conn is not supported when not online")
        else:
            super(TonyDBC, self).close_temp_conn()

    def cursor(self) -> mariadb.Cursor:
        if not self.is_online:
            raise AssertionError("cursor is not supported when not online")
        else:
            return super(TonyDBC, self).cursor()

    def begin_transaction(self) -> None:
        if not self.is_online:
            raise AssertionError("begin_transaction is not supported when not online")
        else:
            super(TonyDBC, self).begin_transaction()

    def commit(self) -> None:
        if not self.is_online:
            raise AssertionError("commit is not supported when not online")
        else:
            super(TonyDBC, self).commit()

    def get_data(
        self,
        query: str,
        before_retry_cmd=None,
        no_tracking=False,
        return_type_codes=False,
    ) -> Any:
        if not self.is_online:
            raise AssertionError("get_data can only be used while online")
        else:
            return super(TonyDBC, self).get_data(
                query=query,
                before_retry_cmd=before_retry_cmd,
                no_tracking=no_tracking,
                return_type_codes=return_type_codes,
            )

    def drop_database(self, database) -> None:
        if not self.is_online:
            raise AssertionError("Cannot drop database if not online")
        else:
            super(TonyDBC, self).drop_database(database)

    def query_table(self, table: str, query: str) -> pd.DataFrame:
        if not self.is_online:
            raise AssertionError("Cannot query table if not online")
        else:
            return super(TonyDBC, self).query_table(table, query)

    def refresh_table(self, table) -> None:
        if not self.is_online:
            raise AssertionError("Cannot refresh table if not online")
        else:
            super(TonyDBC, self).refresh_table(table)

    def update_table(self, table_name, df) -> None:
        """Instead of actually updating the table, just enqueue the updates for doing later."""
        kwargs = {"table_name": table_name, "df": df}
        if self.is_online:
            super(TonyDBC, self).update_table(**kwargs)
        else:
            self.__update_queue.put(("update_table", kwargs))

    def write_dataframe(self, df, table_name, if_exists="replace", index=False) -> None:
        kwargs = {
            "df": df,
            "table_name": table_name,
            "if_exists": if_exists,
            "index": index,
        }
        if self.is_online:
            super(TonyDBC, self).write_dataframe(**kwargs)
        else:
            self.__update_queue.put(("write_dataframe", kwargs))

    def update_blob(
        self, table_name, blob_column, id_value, filepath, max_size_MB=16
    ) -> None:
        kwargs = {
            "table_name": table_name,
            "blob_column": blob_column,
            "id_value": id_value,
            "filepath": filepath,
            "max_size_MB": max_size_MB,
        }
        if self.is_online:
            super(TonyDBC, self).update_blob(**kwargs)
        else:
            self.__update_queue.put(("update_blob", kwargs))

    def post_data(self, query: str) -> Any:
        kwargs = {"query": query}
        if self.is_online:
            return super(TonyDBC, self).post_data(**kwargs)
        else:
            self.__update_queue.put(("post_data", kwargs))

    def post_datalist(self, query: str, insert_data: list) -> Any:
        kwargs = {
            "query": query,
            "insert_data": insert_data,
        }
        if self.is_online:
            return super(TonyDBC, self).post_datalist(
                query=query, insert_data=insert_data
            )
        else:
            self.__update_queue.put(("post_datalist", kwargs))

    def execute(
        self,
        command: str,
        command_values=None,
        before_retry_cmd=None,
        no_tracking=False,
        log_progress=False,
    ) -> None:
        kwargs = {
            "command": command,
            "command_values": command_values,
            "before_retry_cmd": before_retry_cmd,
            "no_tracking": no_tracking,
            "log_progress": log_progress,
        }
        if self.is_online:
            super(TonyDBC, self).execute(**kwargs)
        else:
            self.__update_queue.put(("execute", kwargs))

    def execute_script(
        self, script_path: str, get_return_values=False, cur_database=None
    ) -> Any:
        kwargs = {
            "script_path": script_path,
            "get_return_values": get_return_values,
            "cur_database": cur_database,
        }
        if self.is_online:
            return super(TonyDBC, self).execute_script(**kwargs)
        else:
            self.__update_queue.put(("execute_script", kwargs))

    def insert_row_all_string(self, table, row_dict) -> None:
        kwargs = {
            "table": table,
            "row_dict": row_dict,
        }

        if self.is_online:
            super(TonyDBC, self).insert_row_all_string(**kwargs)
        else:
            self.__update_queue.put(("insert_row_all_string", kwargs))

    def append_to_table(
        self, table, df, return_reindexed=False, no_tracking=False
    ) -> pd.DataFrame | None:
        kwargs = {
            "table": table,
            "df": df,
            "return_reindexed": return_reindexed,
            "no_tracking": no_tracking,
        }
        if self.is_online:
            return super(TonyDBC, self).append_to_table(**kwargs)
        else:
            if return_reindexed:
                raise AssertionError(
                    "TonyDBC.append_to_table: cannot return reindexed while in `offline` mode"
                )
            self.__update_queue.put(("append_to_table", kwargs))
            return None

    def log_to_db(self, log_dict: dict) -> None:
        kwargs = {"log_dict": log_dict}
        if self.is_online:
            super(TonyDBC, self).log_to_db(**kwargs)
        else:
            self.__update_queue.put(("log_to_db", kwargs))



"""
Comprehensive unit tests for EVERY method in the TonyDBC class.

This test suite provides complete coverage of all TonyDBC methods using mocks
to avoid requiring actual database connections for unit testing.
"""

import os
import pickle
import sys
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch

import mariadb  # type: ignore
import pandas as pd
import pytest
from mariadb.constants.CLIENT import MULTI_STATEMENTS  # type: ignore

# Set required environment variables BEFORE importing tonydbc
os.environ.setdefault("USE_PRODUCTION_DATABASE", "False")
os.environ.setdefault("CHECK_ENVIRONMENT_INTEGRITY", "False")
os.environ.setdefault("INTERACT_AFTER_ERROR", "False")
os.environ.setdefault("DEFAULT_TIMEZONE", "UTC")
os.environ.setdefault("MYSQL_DATABASE", "test")
os.environ.setdefault("MYSQL_HOST", "localhost")
os.environ.setdefault("MYSQL_READWRITE_USER", "test")
os.environ.setdefault("MYSQL_READWRITE_PASSWORD", "test")
os.environ.setdefault("MYSQL_PRODUCTION_DATABASE", "test_prod")
os.environ.setdefault("MYSQL_TEST_DATABASE", "test")

# Add the src directory to the path
sys.path.insert(0, str(Path(__file__).parent.parent / "src"))

from tonydbc import TonyDBC
from tonydbc.tonydbc import _TonyDBCOnlineOnly


class TestTonyDBCOnlineOnly:
    """Test suite for _TonyDBCOnlineOnly class methods"""

    @pytest.fixture
    def mock_connection(self):
        """Create a mock mariadb connection"""
        mock_conn = Mock(spec=mariadb.Connection)
        mock_cursor = Mock(spec=mariadb.Cursor)

        # Set up default return values for cursor methods
        mock_cursor.fetchall.return_value = []
        mock_cursor.description = []
        mock_cursor.execute.return_value = None
        mock_cursor.executemany.return_value = None
        mock_cursor.lastrowid = 1

        # Properly mock the context manager behavior
        cursor_context_manager = Mock()
        cursor_context_manager.__enter__ = Mock(return_value=mock_cursor)
        cursor_context_manager.__exit__ = Mock(return_value=None)
        mock_conn.cursor.return_value = cursor_context_manager

        # Set up connection methods
        mock_conn.ping.return_value = None
        mock_conn.commit.return_value = None
        mock_conn.begin.return_value = None
        mock_conn.close.return_value = None
        mock_conn._closed = False

        return mock_conn, mock_cursor

    @pytest.fixture
    def tonydbc_instance(self, mock_connection):
        """Create a TonyDBC instance with mocked connection and common test data"""
        mock_conn, mock_cursor = mock_connection

        # Create the instance without calling __enter__
        db = _TonyDBCOnlineOnly(
            host="localhost", user="test", password="test", database="test", port=3306
        )
        # Manually set the connection to avoid calling make_connection
        db._mariatonydbcn = mock_conn

        # Set up common attributes that would normally be set by set_timezone
        db.session_timezone = "UTC"
        import dateutil.tz

        db.default_tz = dateutil.tz.gettz("UTC")

        # Set up common test data structures
        db._primary_keys = {
            "test_table": "id",
            "users": "user_id",
            "products": "product_id",
        }
        db._column_datatypes = {
            "test_table": {"id": int, "name": str, "email": str},
            "users": {"user_id": int, "username": str, "email": str},
            "products": {"product_id": int, "name": str, "price": float},
        }

        yield db, mock_conn, mock_cursor

    def test_init_with_all_parameters(self):
        """Test __init__ with all parameters provided"""
        db = _TonyDBCOnlineOnly(
            host="testhost",
            user="testuser",
            password="testpass",
            database="testdb",
            port=3307,
            media_to_deserialize={"test_table": ["col1", "col2"]},
            autocommit=False,
            logger_instance=Mock(),
            prefix="TEST",
            lost_connection_callback=Mock(),
            session_timezone="Asia/Bangkok",
            interact_after_error=True,
            force_no_audit=True,
        )

        assert db.host == "testhost"
        assert db.user == "testuser"
        assert db.password == "testpass"
        assert db.database == "testdb"
        assert db.port == 3307
        assert db.media_to_deserialize == {"test_table": ["col1", "col2"]}
        assert not db.autocommit
        assert db.prefix == "TEST"
        assert db.interact_after_error

    def test_init_with_environment_variables(self):
        """Test __init__ using environment variables for credentials"""
        with patch.dict(
            os.environ,
            {
                "MYSQL_HOST": "env_host",
                "MYSQL_READWRITE_USER": "env_user",
                "MYSQL_READWRITE_PASSWORD": "env_pass",
                "MYSQL_DATABASE": "env_db",
            },
        ):
            db = _TonyDBCOnlineOnly()
            assert db.host == "env_host"
            assert db.user == "env_user"
            assert db.password == "env_pass"
            assert db.database == "env_db"

    def test_init_missing_credentials_raises_error(self):
        """Test __init__ raises error when credentials are missing"""
        with patch.dict(os.environ, {}, clear=True):
            with pytest.raises(AssertionError, match="Not all credentials provided"):
                _TonyDBCOnlineOnly()

    def test_enter_returns_self(self, tonydbc_instance):
        """Test __enter__ returns self (bug has been fixed)"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Disable auditing to avoid audit connection setup
        db.do_audit = False

        # Mock the mariadb.connect call and timezone query
        with (
            patch("tonydbc.tonydbc.mariadb.connect", return_value=mock_conn),
            patch.object(db, "get_data") as mock_get_data,
        ):
            mock_get_data.return_value = [{"@@session.time_zone": "+00:00"}]
            result = db.__enter__()

        # Now correctly returns self after fixing the bug
        assert result is db

    @patch("tonydbc.tonydbc.mariadb.connect")
    def test_make_connection_success(self, mock_connect, tonydbc_instance):
        """Test make_connection establishes connection successfully"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_connect.return_value = mock_conn

        # Mock the timezone query that set_timezone calls
        with patch.object(db, "get_data") as mock_get_data:
            mock_get_data.return_value = [{"@@session.time_zone": "+00:00"}]

            db.make_connection()

        mock_connect.assert_called_once_with(
            host=db.host,
            port=db.port,
            user=db.user,
            password=db.password,
            database=db.database,
            client_flag=MULTI_STATEMENTS,
            autocommit=db.autocommit,
            read_timeout=3600,
            write_timeout=3600,
            local_infile=True,
            compress=True,
        )

    @patch("tonydbc.tonydbc.mariadb.connect")
    def test_make_connection_retry_on_failure(self, mock_connect, tonydbc_instance):
        """Test make_connection retries on connection failure"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_connect.side_effect = [
            mariadb.OperationalError("Connection failed"),
            mariadb.OperationalError("Connection failed"),
            mock_conn,
        ]

        # Mock the timezone query that set_timezone calls after successful connection
        with patch.object(db, "get_data") as mock_get_data:
            mock_get_data.return_value = [{"@@session.time_zone": "+00:00"}]

            db.make_connection()

        assert mock_connect.call_count == 3

    def test_set_timezone_with_parameter(self, tonydbc_instance):
        """Test set_timezone with timezone parameter"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Mock the timezone query that set_timezone calls
        with (
            patch.object(db, "get_data") as mock_get_data,
            patch.dict(os.environ, {"DEFAULT_TIMEZONE": "UTC"}),
        ):
            mock_get_data.return_value = [{"@@session.time_zone": "+07:00"}]

            db.set_timezone("Asia/Bangkok")

        assert db.session_timezone == "Asia/Bangkok"

    def test_set_timezone_with_default(self, tonydbc_instance):
        """Test set_timezone using DEFAULT_TIMEZONE environment variable"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Mock the timezone query that set_timezone calls
        with (
            patch.object(db, "get_data") as mock_get_data,
            patch.dict(os.environ, {"DEFAULT_TIMEZONE": "Europe/London"}),
        ):
            mock_get_data.return_value = [{"@@session.time_zone": "+01:00"}]

            db.set_timezone()

        assert db.session_timezone == "Europe/London"

    def test_exit_commits_and_closes_connection(self, tonydbc_instance):
        """Test __exit__ commits transactions and closes connection"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.autocommit = False

        db.__exit__(None, None, None)

        mock_conn.close.assert_called_once()

    def test_exit_handles_closed_connection(self, tonydbc_instance):
        """Test __exit__ handles already closed connection gracefully"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_conn._closed = True

        db.__exit__(None, None, None)

        mock_conn.close.assert_not_called()

    def test_start_temp_conn(self, tonydbc_instance):
        """Test start_temp_conn creates temporary connection"""
        db, mock_conn, mock_cursor = tonydbc_instance
        original_conn = mock_conn

        with patch.object(db, "__enter__"):
            db.start_temp_conn()

        assert db.using_temp_conn
        assert db._mariatonydbcn_old == original_conn

    def test_start_temp_conn_already_using_temp_raises_error(self, tonydbc_instance):
        """Test start_temp_conn raises error if already using temp connection"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.using_temp_conn = True

        with pytest.raises(
            AssertionError, match="already using your temporary connection"
        ):
            db.start_temp_conn()

    def test_close_temp_conn(self, tonydbc_instance):
        """Test close_temp_conn restores original connection"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.using_temp_conn = True
        original_conn = Mock()
        db._mariatonydbcn_old = original_conn

        with patch.object(db, "__exit__"):
            db.close_temp_conn()

        assert not db.using_temp_conn
        assert db._mariatonydbcn == original_conn

    def test_close_temp_conn_not_using_temp_raises_error(self, tonydbc_instance):
        """Test close_temp_conn raises error if not using temp connection"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.using_temp_conn = False

        with pytest.raises(AssertionError, match="not using your temporary connection"):
            db.close_temp_conn()

    def test_now_returns_timestamp(self, tonydbc_instance):
        """Test now() returns current timestamp in session timezone"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.session_timezone = "UTC"

        result = db.now()

        assert isinstance(result, pd.Timestamp)
        # Check timezone - could be datetime.timezone.utc or pytz timezone
        assert result.tz is not None
        assert str(result.tz) in ["UTC", "UTC+00:00", "+00:00"]

    def test_cursor_returns_connection_cursor(self, tonydbc_instance):
        """Test cursor() returns database cursor"""
        db, mock_conn, mock_cursor = tonydbc_instance

        db.cursor()

        mock_conn.cursor.assert_called_once()

    def test_begin_transaction_disables_autocommit(self, tonydbc_instance):
        """Test begin_transaction disables autocommit and begins transaction"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.autocommit = True

        with patch.object(db, "execute") as mock_execute:
            db.begin_transaction()

        mock_execute.assert_called_once_with("SET AUTOCOMMIT=0;")
        mock_conn.begin.assert_called_once()

    def test_begin_transaction_already_disabled_autocommit(self, tonydbc_instance):
        """Test begin_transaction when autocommit already disabled"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.autocommit = False

        with patch.object(db, "execute") as mock_execute:
            db.begin_transaction()

        mock_execute.assert_not_called()
        mock_conn.begin.assert_called_once()

    def test_commit_success(self, tonydbc_instance):
        """Test commit() commits transaction successfully"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.autocommit = True

        with patch.object(db, "execute") as mock_execute:
            db.commit()

        mock_conn.commit.assert_called_once()
        mock_execute.assert_called_once_with("SET AUTOCOMMIT=1;")

    def test_commit_interface_error_with_callback(self, tonydbc_instance):
        """Test commit() handles InterfaceError and calls callback"""
        db, mock_conn, mock_cursor = tonydbc_instance
        callback = Mock()
        db._lost_connection_callback = callback
        mock_conn.commit.side_effect = mariadb.InterfaceError("Lost connection")

        with pytest.raises(ConnectionError, match="Database server is down"):
            db.commit()

        callback.assert_called_once()

    def test_log_with_logger(self, tonydbc_instance):
        """Test log() uses provided logger instance"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_logger = Mock()
        db._l = mock_logger

        db.log("test message")

        mock_logger.info.assert_called_once_with("test message")

    def test_log_with_prefix(self, tonydbc_instance):
        """Test log() adds prefix to message"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.prefix = "TEST"
        mock_logger = Mock()
        db._l = mock_logger

        db.log("message")

        mock_logger.info.assert_called_once_with("TEST | message")

    def test_log_without_logger_prints(self, tonydbc_instance):
        """Test log() prints to stdout when no logger provided"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db._l = None

        with patch("builtins.print") as mock_print:
            db.log("test message")

        mock_print.assert_called_once_with("test message")

    def test_column_info_all_tables(self, tonydbc_instance):
        """Test column_info() returns info for all tables"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_data = [{"COLUMN_NAME": "id", "DATA_TYPE": "int"}]

        with patch.object(db, "get_data", return_value=mock_data):
            result = db.column_info()

        assert isinstance(result, pd.DataFrame)
        assert len(result) == 1

    def test_column_info_specific_table(self, tonydbc_instance):
        """Test column_info() returns info for specific table"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_data = [{"COLUMN_NAME": "name", "DATA_TYPE": "varchar"}]

        with patch.object(db, "get_data", return_value=mock_data) as mock_get_data:
            db.column_info("users")

        mock_get_data.assert_called_once()
        query = mock_get_data.call_args[0][0]
        assert "TABLE_NAME = 'users'" in query

    def test_last_insert_id_returns_id(self, tonydbc_instance):
        """Test last_insert_id property returns last inserted ID"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_data = [{"id": 123}]

        with patch.object(db, "get_data", return_value=mock_data):
            result = db.last_insert_id

        assert result == 123

    def test_use_changes_database(self, tonydbc_instance):
        """Test use() changes to new database"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with (
            patch.object(db, "__exit__") as mock_exit,
            patch.object(db, "__enter__") as mock_enter,
        ):
            db.use("new_database")

        assert db.database == "new_database"
        mock_exit.assert_called_once_with(None, None, None)
        mock_enter.assert_called_once()

    def test_iso_timestamp_to_session_time_zone(self, tonydbc_instance):
        """Test iso_timestamp_to_session_time_zone converts timestamp"""
        db, mock_conn, mock_cursor = tonydbc_instance

        result = db.iso_timestamp_to_session_time_zone("2023-01-01T12:00:00+00:00")

        assert isinstance(result, str)
        assert "2023-01-01" in result

    @pytest.mark.skip(
        reason="Replaced by integration test that checks persisted results"
    )
    def test_write_dataframe(self, tonydbc_instance):
        pass

    def test_update_table(self, tonydbc_instance):
        """Test update_table performs UPDATE operations"""
        db, mock_conn, mock_cursor = tonydbc_instance
        df = pd.DataFrame({"id": [1], "name": ["Updated"]})

        db.update_table("test_table", df)

        # Verify that executemany was called on the cursor
        mock_cursor.executemany.assert_called_once()
        # Check that the UPDATE command was constructed correctly
        call_args = mock_cursor.executemany.call_args
        assert "UPDATE test_table" in call_args[0][0]
        assert "SET id = ?, name = ?" in call_args[0][0]
        assert "WHERE `id` = ?" in call_args[0][0]

    def test_update_blob_success(self, tonydbc_instance):
        """Test update_blob uploads file to BLOB column"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(b"test data")
            tmp_file.flush()

            with patch.object(db, "execute") as mock_execute:
                db.update_blob("test_table", "blob_col", 1, tmp_file.name)

            mock_execute.assert_called()

        os.unlink(tmp_file.name)

    def test_update_blob_file_too_large(self, tonydbc_instance):
        """Test update_blob raises error for oversized file"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(b"x" * (17 * 1024 * 1024))  # 17MB file
            tmp_file.flush()

            with pytest.raises(AssertionError, match="too large"):
                db.update_blob(
                    "test_table", "blob_col", 1, tmp_file.name, max_size_MB=16
                )

        os.unlink(tmp_file.name)

    def test_read_dataframe_from_table(self, tonydbc_instance):
        """Test read_dataframe_from_table returns DataFrame"""
        pytest.skip("Covered by integration tests using real tables")

    def test_get_data_success(self, tonydbc_instance):
        """Test get_data executes query and returns data"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_cursor.fetchall.return_value = [(1, "test")]
        mock_cursor.description = [("id",), ("name",)]

        result = db.get_data("SELECT * FROM test_table")

        assert isinstance(result, list)
        mock_cursor.execute.assert_called_once_with("SELECT * FROM test_table")

    def test_get_data_with_retry_command(self, tonydbc_instance):
        """Test get_data executes before_retry_cmd on retry attempts"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Create a call counter to track which call we're on
        call_count = 0

        def mock_execute_side_effect(query):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                # First attempt fails
                raise mariadb.InterfaceError("Connection lost")
            # Subsequent calls succeed
            return None

        mock_cursor.execute.side_effect = mock_execute_side_effect
        mock_cursor.fetchall.return_value = []
        mock_cursor.description = []

        with patch.object(db, "make_connection"):  # Mock reconnection
            db.get_data(
                "SELECT * FROM test", before_retry_cmd="SET SESSION sql_mode = ''"
            )

        # Verify that the retry command was executed at some point
        call_args_list = mock_cursor.execute.call_args_list
        retry_commands = [
            call[0][0]
            for call in call_args_list
            if call[0][0] == "SET SESSION sql_mode = ''"
        ]
        assert len(retry_commands) == 1, (
            f"Expected retry command to be called once, but got: {[call[0][0] for call in call_args_list]}"
        )

        # Verify that the main query was attempted multiple times
        main_queries = [
            call[0][0] for call in call_args_list if call[0][0] == "SELECT * FROM test"
        ]
        assert len(main_queries) >= 2, "Expected main query to be retried"

    def test_databases_property(self, tonydbc_instance):
        """Test databases property returns list of databases"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_data = [{"Database": "db1"}, {"Database": "db2"}]

        with patch.object(db, "get_data", return_value=mock_data):
            result = db.databases

        assert result == ["db1", "db2"]

    def test_users_property(self, tonydbc_instance):
        """Test users property returns list of (host, user) tuples"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_data = [{"User": "user1", "Host": "host1"}]

        with patch.object(db, "get_data", return_value=mock_data):
            result = db.users

        assert len(result) == 1
        assert result[0] == ("host1", "user1")  # (Host, User) tuple

    def test_show_grants(self, tonydbc_instance):
        """Test show_grants returns user grants"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_data = [{"Grants": "GRANT SELECT ON *.* TO 'user'@'%'"}]

        with patch.object(db, "get_data", return_value=mock_data):
            result = db.show_grants("testuser")

        assert result == mock_data

    def test_production_databases_from_env(self, tonydbc_instance):
        """Test production_databases property reads from environment and adds system DBs"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with patch.dict("os.environ", {"PRODUCTION_DATABASES": '["prod1", "prod2"]'}):
            result = db.production_databases

        # Should include env databases + system databases, sorted
        expected = sorted(
            [
                "prod1",
                "prod2",
                "information_schema",
                "mysql",
                "performance_schema",
                "sys",
            ]
        )
        assert result == expected

    def test_production_databases_empty_list(self, tonydbc_instance):
        """Test production_databases returns only system DBs when env var not set"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with patch.dict("os.environ", {}, clear=True):
            result = db.production_databases

        # Should only include system databases when no env var is set
        expected = sorted(["information_schema", "mysql", "performance_schema", "sys"])
        assert result == expected

    def test_drop_database_production_raises_error(self, tonydbc_instance):
        """Test drop_database raises error for production database"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Set the private attribute directly to simulate production databases
        db._TonyDBCOnlineOnly__production_databases = ["prod_db"]

        with pytest.raises(AssertionError, match="production database"):
            db.drop_database("prod_db")

    def test_drop_database_success(self, tonydbc_instance):
        """Test drop_database executes DROP DATABASE command"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Set the private attribute to empty list (no production databases)
        db._TonyDBCOnlineOnly__production_databases = []

        def mock_get_data(query):
            if query == "SHOW DATABASES;":
                return [{"Database": "test_db"}, {"Database": "other_db"}]
            return []

        with (
            patch.object(db, "get_data", side_effect=mock_get_data),
            patch.object(db, "execute") as mock_execute,
        ):
            db.drop_database("test_db")

        # Should call execute twice: UNLOCK TABLES and DROP DATABASE
        assert mock_execute.call_count == 2
        mock_execute.assert_any_call("USE test_db; UNLOCK TABLES;")
        mock_execute.assert_any_call("DROP DATABASE test_db;")

    def test_post_data(self, tonydbc_instance):
        """Test post_data executes query"""
        db, mock_conn, mock_cursor = tonydbc_instance

        db.post_data("INSERT INTO test VALUES (1)")

        mock_cursor.execute.assert_called_once_with("INSERT INTO test VALUES (1)")

    def test_post_datalist(self, tonydbc_instance):
        """Test post_datalist executes batch query"""
        db, mock_conn, mock_cursor = tonydbc_instance
        data = [(1, "A"), (2, "B")]

        db.post_datalist("INSERT INTO test VALUES (?, ?)", data)

        mock_cursor.executemany.assert_called_once_with(
            "INSERT INTO test VALUES (?, ?)", data
        )

    def test_execute_simple_command(self, tonydbc_instance):
        """Test execute runs simple SQL command"""
        db, mock_conn, mock_cursor = tonydbc_instance

        db.execute("CREATE TABLE test (id INT)")

        mock_cursor.execute.assert_called_once_with("CREATE TABLE test (id INT)")

    def test_execute_with_values(self, tonydbc_instance):
        """Test execute runs command with parameter values"""
        db, mock_conn, mock_cursor = tonydbc_instance

        db.execute("INSERT INTO test VALUES (?)", (123,))

        mock_cursor.execute.assert_called_once_with(
            "INSERT INTO test VALUES (?)", (123,)
        )

    def test_execute_with_progress_logging(self, tonydbc_instance):
        """Test execute with progress logging enabled"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with patch.object(db, "log") as mock_log:
            db.execute("LONG RUNNING QUERY", log_progress=True)

        mock_log.assert_called()

    def test_execute_script_success(self, tonydbc_instance):
        """Test execute_script runs SQL script file"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".sql", delete=False
        ) as tmp_file:
            tmp_file.write("CREATE TABLE test (id INT);")
            tmp_file.flush()

            db.execute_script(tmp_file.name)

            # Should call cursor.execute with the SQL command
            mock_cursor.execute.assert_called_once_with("CREATE TABLE test (id INT);")

        os.unlink(tmp_file.name)

    def test_execute_script_with_return_values(self, tonydbc_instance):
        """Test execute_script with return values enabled"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Mock cursor to return data
        mock_cursor.fetchall.return_value = [(1,)]
        mock_cursor.description = [
            ("result", None, None, None, None, None, None, None, None, None, None)
        ]

        with tempfile.NamedTemporaryFile(
            mode="w", suffix=".sql", delete=False
        ) as tmp_file:
            tmp_file.write("SELECT 1 as result;")
            tmp_file.flush()

            result = db.execute_script(tmp_file.name, get_return_values=True)

            # Should return list with one dict containing the result
            assert len(result) == 1
            assert result[0] == [{"result": 1}]

        os.unlink(tmp_file.name)

    def test_get_primary_key_cached(self, tonydbc_instance):
        """Test get_primary_key returns cached primary key"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db._primary_keys = {"test_table": "id"}

        result = db.get_primary_key("test_table")

        assert result == "id"

    def test_get_primary_key_with_default(self, tonydbc_instance):
        """Test get_primary_key returns default when table has no primary key"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Create an actual table with no primary key in the mock database
        # First, create the table structure in our mock data
        db._primary_keys = {
            "users": "user_id",  # Other tables have primary keys
        }

        # Mock get_data to simulate a table that exists but has no primary key
        def mock_get_data(query, no_tracking=False):
            if "DESCRIBE no_pk_table" in query:
                # Return table description with columns but no primary key (Key is empty)
                return [
                    {
                        "Field": "col1",
                        "Type": "varchar(100)",
                        "Null": "YES",
                        "Key": "",
                        "Default": None,
                        "Extra": "",
                    },
                    {
                        "Field": "col2",
                        "Type": "int(11)",
                        "Null": "YES",
                        "Key": "",
                        "Default": None,
                        "Extra": "",
                    },
                ]
            return []

        with patch.object(db, "get_data", side_effect=mock_get_data):
            with patch.object(
                db, "refresh_primary_keys"
            ):  # Mock to avoid real DB calls
                result = db.get_primary_key("no_pk_table", default="pk_id")

        assert result == "pk_id"

    def test_refresh_primary_keys(self, tonydbc_instance):
        """Test refresh_primary_keys loads primary key information"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_data = [{"TABLE_NAME": "test", "COLUMN_NAME": "id"}]

        with patch.object(db, "get_data", return_value=mock_data):
            db.refresh_primary_keys()

        assert db._primary_keys == {"test": "id"}

    def test_primary_keys_property(self, tonydbc_instance):
        """Test primary_keys property returns primary key dict"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with patch.object(db, "refresh_primary_keys"):
            db._primary_keys = {"table1": "id", "table2": "pk"}
            result = db.primary_keys

        assert result == {"table1": "id", "table2": "pk"}

    def test_get_column_datatypes_cached(self, tonydbc_instance):
        """Test get_column_datatypes returns cached datatypes"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db._column_datatypes = {"test_table": {"id": int, "name": str}}

        result = db.get_column_datatypes("test_table")

        assert result == {"id": int, "name": str}

    def test_get_column_datatypes_describe_table(self, tonydbc_instance):
        """Test get_column_datatypes uses DESCRIBE when not cached"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db._column_datatypes = {}
        mock_data = [{"Field": "id", "Type": "int(11)"}]

        with patch.object(db, "get_data", return_value=mock_data):
            result = db.get_column_datatypes("test_table")

        assert "id" in result

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_column_datatypes_property(self, tonydbc_instance):
        """Test column_datatypes property returns datatype dict"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with patch.object(db, "refresh_column_datatypes"):
            db._column_datatypes = {"table1": {"id": int}}
            result = db.column_datatypes

        assert result == {"table1": {"id": int}}

    def test_non_primary_keys(self, tonydbc_instance):
        """Test non_primary_keys returns non-primary key columns"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with (
            patch.object(
                db,
                "get_column_datatypes",
                return_value={"id": int, "name": str, "email": str},
            ),
            patch.object(db, "get_primary_key", return_value="id"),
        ):
            result = db.non_primary_keys("test_table")

        assert result == ["name", "email"]

    def test_connection_params(self, tonydbc_instance):
        """Test connection_params returns connection parameters"""
        db, mock_conn, mock_cursor = tonydbc_instance

        result = db.connection_params

        assert result["host"] == db.host
        assert result["user"] == db.user
        assert result["password"] == db.password
        assert result["database"] == db.database
        assert result["port"] == db.port

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_insert_row_all_string(self, tonydbc_instance):
        """Test insert_row_all_string inserts row with string values"""
        db, mock_conn, mock_cursor = tonydbc_instance
        row_data = {"id": "1", "name": "test"}

        with patch.object(db, "execute") as mock_execute:
            db.insert_row_all_string("test_table", row_data)

        mock_execute.assert_called_once()
        query = mock_execute.call_args[0][0]
        assert "INSERT INTO test_table" in query

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_temp_id_table(self, tonydbc_instance):
        """Test temp_id_table creates temporary table with IDs"""
        db, mock_conn, mock_cursor = tonydbc_instance
        id_list = [1, 2, 3]

        with patch.object(db, "execute") as mock_execute:
            context_manager = db.temp_id_table(id_list)
            with context_manager as temp_table_name:
                assert temp_table_name.startswith("temp_ids_")

        # Verify table creation and cleanup
        assert mock_execute.call_count >= 2  # CREATE and DROP

    @pytest.mark.skip(reason="Replaced by integration test with real table")
    def test_append_to_table_basic(self, tonydbc_instance):
        pass

    def test_append_to_table_with_reindexing(self, tonydbc_instance):
        """Test append_to_table with return_reindexed=True"""
        pytest.skip("Replaced by integration test asserting reindexed results")

    def test_query_table_basic(self, tonydbc_instance):
        """Test query_table returns DataFrame from table"""
        pytest.skip("Covered by integration test using explicit SQL on real table")

    def test_query_table_with_custom_query(self, tonydbc_instance):
        """Test query_table with custom SQL query"""
        pytest.skip("Covered by integration test using explicit SQL on real table")

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_refresh_table(self, tonydbc_instance):
        """Test refresh_table loads entire table"""
        db, mock_conn, mock_cursor = tonydbc_instance
        mock_df = pd.DataFrame({"id": [1, 2, 3]})

        with patch.object(db, "query_table", return_value=mock_df):
            result = db.refresh_table("test_table")

        assert isinstance(result, pd.DataFrame)

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_log_to_db(self, tonydbc_instance):
        """Test log_to_db saves log entry to database"""
        db, mock_conn, mock_cursor = tonydbc_instance
        log_data = {"level": "INFO", "message": "test log"}

        with patch.object(db, "execute") as mock_execute:
            db.log_to_db(log_data)

        mock_execute.assert_called_once()

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_save_instrumentation_to_file(self, tonydbc_instance):
        """Test _save_instrumentation saves to CSV file"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.do_audit = True
        db.ipath = Path("test_audit.csv")

        with patch("pandas.DataFrame.to_csv") as mock_to_csv:
            db._save_instrumentation(
                method="test_method",
                table="test_table",
                query="SELECT 1",
                started_at=pd.Timestamp.now(),
                payload_size=100,
                num_rows=1,
                num_cols=1,
            )

        mock_to_csv.assert_called_once()

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_save_instrumentation_to_database(self, tonydbc_instance):
        """Test _save_instrumentation saves to audit database"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.do_audit = True
        db.ipath = "database"
        db._audit_db = Mock()

        db._save_instrumentation(
            method="test_method",
            table="test_table",
            query="SELECT 1",
            started_at=pd.Timestamp.now(),
            payload_size=100,
            num_rows=1,
            num_cols=1,
        )

        db._audit_db.execute.assert_called_once()


class TestTonyDBC:
    """Test suite for TonyDBC class methods (offline-capable version)"""

    @pytest.fixture
    def mock_connection(self):
        """Create a mock mariadb connection"""
        mock_conn = Mock(spec=mariadb.Connection)
        mock_cursor = Mock(spec=mariadb.Cursor)

        # Set up default return values for cursor methods
        mock_cursor.fetchall.return_value = []
        mock_cursor.description = []
        mock_cursor.execute.return_value = None
        mock_cursor.executemany.return_value = None
        mock_cursor.lastrowid = 1

        # Properly mock the context manager behavior
        cursor_context_manager = Mock()
        cursor_context_manager.__enter__ = Mock(return_value=mock_cursor)
        cursor_context_manager.__exit__ = Mock(return_value=None)
        mock_conn.cursor.return_value = cursor_context_manager

        # Set up connection methods
        mock_conn.ping.return_value = None
        mock_conn.commit.return_value = None
        mock_conn.begin.return_value = None
        mock_conn.close.return_value = None
        mock_conn._closed = False

        return mock_conn, mock_cursor

    @pytest.fixture
    def tonydbc_instance(self, mock_connection):
        """Create a TonyDBC instance with mocked connection"""
        mock_conn, mock_cursor = mock_connection

        with patch.dict(os.environ, {"MEDIA_BASE_PATH_PRODUCTION": "/tmp"}):
            # Create the instance without calling __enter__
            db = TonyDBC(
                host="localhost",
                user="test",
                password="test",
                database="test",
                port=3306,
            )
            # Manually set the connection to avoid calling make_connection
            db._mariatonydbcn = mock_conn
            yield db, mock_conn, mock_cursor

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_init_sets_media_base_path(self, mock_connection):
        """Test TonyDBC __init__ sets media base path from environment"""
        mock_conn, mock_cursor = mock_connection

        with (
            patch("tonydbc.tonydbc.mariadb.connect", return_value=mock_conn),
            patch.dict(os.environ, {"MEDIA_BASE_PATH_PRODUCTION": "/custom/path"}),
        ):
            db = TonyDBC(
                host="localhost", user="test", password="test", database="test"
            )

        assert db.media_base_path == "/custom/path"

    def test_is_online_property_getter(self, tonydbc_instance):
        """Test is_online property getter"""
        db, mock_conn, mock_cursor = tonydbc_instance

        assert db.is_online  # Default state

    def test_is_online_property_setter_to_false(self, tonydbc_instance):
        """Test is_online property setter to False"""
        db, mock_conn, mock_cursor = tonydbc_instance

        db.is_online = False

        assert db._TonyDBC__offline_status == "offline"

    def test_is_online_property_setter_to_true_flushes_updates(self, tonydbc_instance):
        """Test is_online property setter to True flushes queued updates"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db._TonyDBC__offline_status = "offline"

        with patch.object(db, "flush_updates") as mock_flush:
            db.is_online = True

        mock_flush.assert_called_once()

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_flush_updates_processes_queue(self, tonydbc_instance):
        """Test flush_updates processes all queued operations"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Add some operations to the queue
        db._TonyDBC__update_queue.put(
            ("execute", {"command": "CREATE TABLE test (id INT)"})
        )
        db._TonyDBC__update_queue.put(
            ("post_data", {"query": "INSERT INTO test VALUES (1)"})
        )

        with (
            patch.object(db, "execute") as mock_execute,
            patch.object(db, "post_data") as mock_post_data,
        ):
            db.flush_updates()

        mock_execute.assert_called_once_with(command="CREATE TABLE test (id INT)")
        mock_post_data.assert_called_once_with(query="INSERT INTO test VALUES (1)")

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_enter_loads_pickled_updates(self, tonydbc_instance):
        """Test __enter__ loads previously pickled updates"""
        db, mock_conn, mock_cursor = tonydbc_instance

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            pickle.dump([("execute", {"command": "SELECT 1"})], tmp_file)
            tmp_file.flush()

            db.pickle_path = tmp_file.name

            with patch.object(db, "flush_updates") as mock_flush:
                db.__enter__()

            mock_flush.assert_called_once()

        os.unlink(tmp_file.name)

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_pickle_updates_saves_queue(self, tonydbc_instance):
        """Test pickle_updates saves queue to file"""
        db, mock_conn, mock_cursor = tonydbc_instance

        # Add operation to queue
        db._TonyDBC__update_queue.put(("execute", {"command": "SELECT 1"}))

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            db.pickle_path = tmp_file.name
            db.pickle_updates()

            # Verify file was created and contains data
            assert os.path.exists(tmp_file.name)

        os.unlink(tmp_file.name)

    def test_exit_pickles_updates_when_offline(self, tonydbc_instance):
        """Test __exit__ pickles updates when offline"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db._TonyDBC__offline_status = "offline"

        with patch.object(db, "pickle_updates") as mock_pickle:
            db.__exit__(None, None, None)

        mock_pickle.assert_called_once()

    def test_offline_methods_queue_operations(self, tonydbc_instance):
        """Test offline methods queue operations instead of executing"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False

        # Test various methods queue operations when offline
        db.execute("CREATE TABLE test (id INT)")
        db.post_data("INSERT INTO test VALUES (1)")
        db.post_datalist("INSERT INTO test VALUES (?)", [(2,), (3,)])

        # Verify operations were queued
        assert db._TonyDBC__update_queue.qsize() == 3

    def test_online_methods_execute_immediately(self, tonydbc_instance):
        """Test online methods execute immediately when online"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = True

        with patch("tonydbc.tonydbc._TonyDBCOnlineOnly.execute") as mock_execute:
            db.execute("CREATE TABLE test (id INT)")

        mock_execute.assert_called_once()

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_offline_read_operations_raise_error(self, tonydbc_instance):
        """Test offline read operations raise appropriate errors"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False

        with pytest.raises(
            AssertionError, match="Cannot read from database while offline"
        ):
            db.get_data("SELECT * FROM test")

        with pytest.raises(
            AssertionError, match="Cannot read from database while offline"
        ):
            db.query_table("test")

        with pytest.raises(
            AssertionError, match="Cannot read from database while offline"
        ):
            db.refresh_table("test")

    @pytest.mark.skip(reason="Temporarily disabled - needs real DB container")
    def test_offline_connection_methods_raise_error(self, tonydbc_instance):
        """Test offline connection methods raise appropriate errors"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False

        with pytest.raises(
            AssertionError, match="Cannot use database connection while offline"
        ):
            db.start_temp_conn()

        with pytest.raises(
            AssertionError, match="Cannot use database connection while offline"
        ):
            db.close_temp_conn()

        with pytest.raises(
            AssertionError, match="Cannot use database connection while offline"
        ):
            db.cursor()

        with pytest.raises(
            AssertionError, match="Cannot use database connection while offline"
        ):
            db.begin_transaction()

        with pytest.raises(
            AssertionError, match="Cannot use database connection while offline"
        ):
            db.commit()

    def test_update_table_queues_when_offline(self, tonydbc_instance):
        """Test update_table queues operation when offline"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False
        df = pd.DataFrame({"id": [1], "name": ["test"]})

        db.update_table("test_table", df)

        assert db._TonyDBC__update_queue.qsize() == 1

    def test_write_dataframe_queues_when_offline(self, tonydbc_instance):
        """Test write_dataframe queues operation when offline"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False
        df = pd.DataFrame({"id": [1], "name": ["test"]})

        db.write_dataframe(df, "test_table")

        assert db._TonyDBC__update_queue.qsize() == 1

    def test_update_blob_queues_when_offline(self, tonydbc_instance):
        """Test update_blob queues operation when offline"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False

        db.update_blob("test_table", "blob_col", 1, "/path/to/file")

        assert db._TonyDBC__update_queue.qsize() == 1

    def test_execute_script_queues_when_offline(self, tonydbc_instance):
        """Test execute_script queues operation when offline"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False

        db.execute_script("/path/to/script.sql")

        assert db._TonyDBC__update_queue.qsize() == 1

    def test_insert_row_all_string_queues_when_offline(self, tonydbc_instance):
        """Test insert_row_all_string queues operation when offline"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False

        db.insert_row_all_string("test_table", {"id": "1", "name": "test"})

        assert db._TonyDBC__update_queue.qsize() == 1

    def test_append_to_table_queues_when_offline(self, tonydbc_instance):
        """Test append_to_table queues operation when offline"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False
        df = pd.DataFrame({"id": [1], "name": ["test"]})

        db.append_to_table("test_table", df)

        assert db._TonyDBC__update_queue.qsize() == 1

    def test_log_to_db_queues_when_offline(self, tonydbc_instance):
        """Test log_to_db queues operation when offline"""
        db, mock_conn, mock_cursor = tonydbc_instance
        db.is_online = False

        db.log_to_db({"level": "INFO", "message": "test"})

        assert db._TonyDBC__update_queue.qsize() == 1


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
Drop-in replacement for pandas.DataFrame methods for reading and writing to database:

    pandas.DataFrame.to_sql_fast
    pandas.read_sql_table

For some reason the current (Jan 2023) implementation in Pandas
using sqlalchemy is really slow.  These methods are ~300x faster.

NOTE: Works only with mariadb's connector for now (pip3 install mariadb)
    https://mariadb.com/resources/blog/how-to-connect-python-programs-to-mariadb/

"""

import code
import csv
import os
import tempfile
from typing import Any, cast

import mariadb  # type: ignore
import numpy as np
import pandas as pd

from .env_utils import get_env_bool

# Map SQL types to Python datatypes
DATATYPE_MAP = {
    "double": np.float64,
    "float": float,
    "int": int,
    "longlong": int,
    "smallint": int,
    "datetime": str,
    "var_string": str,
    "varchar": str,
    "char": str,
    "blob": object,  # Not sure about this one..
    "mediumblob": object,  # Not sure about this one..
    "longblob": object,  # Not sure about this one..
    "bigint": np.int64,
    "shorttext": str,
    "mediumtext": str,
    "longtext": str,
    "enum": str,
    "tinyint": int,
    "text": str,
    "bool": bool,
    "decimal": float,
    "timestamp": str,
    "bit": bool,
    "mediumint": int,
    "integer": int,
    "int8": np.int64,
    "int4": int,
    "int3": int,
    "int2": int,
    "int1": int,
    "middleint": int,
    "serial": int,
}

# TODO: HANDLE ALL THESE:
# 'BIT', 'BLOB', 'DATE', 'DATETIME', 'DATETIME2', 'DECIMAL', 'DOUBLE', 'ENUM',
# 'FLOAT', 'GEOMETRY', 'INT24', 'JSON', 'LONG', 'LONGLONG', 'LONG_BLOB',
# 'MEDIUM_BLOB', 'NEWDATE', 'NEWDECIMAL', 'NULL', 'SET', 'SHORT', 'STRING',
# 'TIME', 'TIME2', 'TIMESTAMP', 'TIMESTAMP2', 'TINY', 'TINY_BLOB', 'VARCHAR',
# 'VAR_STRING', 'YEAR'

# Get a lookup of what all the data_types are
FIELD_TYPE = mariadb.constants.FIELD_TYPE
FIELD_TYPE_DICT = {
    getattr(FIELD_TYPE, k): k for k in dir(FIELD_TYPE) if not k.startswith("_")
}

# To handle as many strings as possible without any issues, let's use some very obscure characters to delimit the CSV
FIELD_DELIMITER = "\x1f"  # Unit Separator (␟)
ENCLOSURE_CHAR = "\x1e"  # Record Separator (␞)
LINE_TERMINATOR = "\n"  # Keep newline as is since the record separator is enough.


class DataFrameFast(pd.DataFrame):
    # Use a distinct name to avoid overriding pandas' final to_sql
    def to_sql_fast(
        self,
        name: str,
        con: mariadb.Connection,
        session_timezone: str,
        if_exists: str = "append",
        index: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if if_exists not in ["replace", "append"]:
            raise AssertionError(
                "not if_exists in ['replace', 'append'] is not yet impemented"
            )

        # Truncate database table
        # NOTE: Users may have to perform to_sql in the correct
        # sequence to avoid causing foreign key errors with this step
        if if_exists == "replace":
            raise AssertionError(
                "'replace' is not implemented correctly... it will erase the whole table!"
            )
            with con.cursor() as cursor:
                r = cursor.execute(f"TRUNCATE TABLE {name}")

        # Nothing to do if the dataframe is empty
        if len(self) == 0:
            return

        assert self.columns.is_unique, (
            f"Dataframe columns are not unique:\n{self.columns}"
        )

        # Drop columns which contain only NULL values since mariadb's stupid executemany might chokes on these
        # if they are in the final column
        # (note that this will screw things up if the column has a DEFAULT of something other than NULL)
        cols_with_all_NULLs = self.columns[pd.isna(self).all()]

        df0 = self.drop(columns=cols_with_all_NULLs).copy()

        # Find any datetime columns.  e.g. dt_cols = {'file_created_at': datetime64[ns, Asia/Bangkok]}
        dt_cols = {
            col: dtype
            for col, dtype in df0.dtypes.to_dict().items()
            if isinstance(dtype, pd.DatetimeTZDtype)
        }

        # Convert all the pandas timestamp columns to string since suddenly we are getting
        # mariadb.NotSupportedError when trying to append
        if len(dt_cols) > 0:
            for dt_col in dt_cols.keys():
                # I think this is okay if we were careful to have the column's data
                # in the current session's datetime format
                new_col = (
                    df0.loc[:, dt_col]
                    # Convert to the correct time zone if necessary
                    .dt.tz_convert(session_timezone)
                    # Convert to a string for mariadb to digest
                    .dt.strftime("%Y-%m-%d %H:%M:%S.%f")
                    .astype("string")
                )

                # Directly assign the converted datetime strings without intermediate assignments
                # This avoids the pandas FutureWarning about incompatible dtype assignment
                # And by creating a fresh Series here, this will work even if df0 is empty
                # Note: we need to create SeriesCtor to satisfy mypy's "Series[Any]" not collable [operator] error
                SeriesCtor = cast(type[pd.Series], pd.Series)
                df0[dt_col] = SeriesCtor(new_col, dtype="string", index=df0.index)

        # Convert np.int64 columns to np.int32 since MySQL cannot accept np.int64 as a param
        small_int64cols = {
            col: dtype
            for col, dtype in df0.dtypes.to_dict().items()
            if isinstance(dtype, np.dtypes.Int64DType)
            # They must fit within the int32 boundaries
            and df0[col].max() < np.iinfo(np.int32).max
            and df0[col].min() > np.iinfo(np.int32).min
        }
        df0 = df0.astype({col: np.int32 for col in small_int64cols})

        # Prepare an INSERT which will populate the real mariadb table with df's data
        # INSERT INTO table(c1,c2,...) VALUES (v11,v12,...), ... (vnn,vn2,...);
        # If index, then we also want the index inserted
        cols: list[str] = ([] if not index else [str(df0.index.name)]) + [
            str(c) for c in list(df0.columns)
        ]
        # Sanitize the column names
        cols = [f"`{c}`" for c in cols]
        assert all(isinstance(c, str) for c in cols)
        cmd = (
            f"INSERT INTO `{name}` ({', '.join([str(c) for c in cols])})"
            f" VALUES ({', '.join(['?'] * len(cols))})"
        )

        # Double-check that our dataframe does not contain our special csv control characters
        def has_ctrl_chars(s: Any) -> bool:
            if isinstance(s, str):
                return ENCLOSURE_CHAR in s or FIELD_DELIMITER in s
            return False

        df_has_ctrl_chars = df0.map(has_ctrl_chars).to_numpy().any()

        if len(df0) == 0:
            # Nothing to INSERT
            return
        elif len(df0) > 3 and not df_has_ctrl_chars:
            temp_dir = tempfile.TemporaryDirectory()
            tmp_filepath = os.path.join(temp_dir.name, name + ".csv")

            # Identify boolean columns and convert them to int (1 for True, 0 for False)
            # to avoid the warning: ('Warning', 1366, "Incorrect integer value:
            # 'False' for column michael_db.cats.is_success at row 1")
            bool_cols = df0.select_dtypes(include=["bool"]).columns
            df0[bool_cols] = df0[bool_cols].astype(int)

            # Convert to csv for uploading to the server as a file, which is faster
            # for some reason
            # the fillna is to handle NULLs
            df0.fillna("\\N").to_csv(
                tmp_filepath,
                index=False,
                header=True,
                quotechar=ENCLOSURE_CHAR,
                quoting=csv.QUOTE_ALL,
                sep=FIELD_DELIMITER,
                lineterminator=LINE_TERMINATOR,
            )

            # Use the faster INFILE method for larger sets of data
            # Set the path to save the CSV file
            tmp_filepath = tmp_filepath.replace("\\", "\\\\")

            # Write DataFrame to a CSV file with specified delimiters and text qualifiers
            cmd = f"""
                LOAD DATA LOCAL INFILE %s
                    INTO TABLE `{name}`
                    FIELDS TERMINATED BY %s
                    ENCLOSED BY %s
                    LINES TERMINATED BY %s
                    IGNORE 1 ROWS
                    ({", ".join([str(c) for c in cols])});
                """

            with con.cursor() as cursor:
                # Perform the insert
                cursor.execute(
                    cmd,
                    (tmp_filepath, FIELD_DELIMITER, ENCLOSURE_CHAR, LINE_TERMINATOR),
                )
                cursor.execute("SHOW WARNINGS;")
                # Check for warnings
                warnings = cursor.fetchall()

            if warnings:
                print(f"TonyDBC ran the command:\n{cmd}\nWARNINGS:\n")
                for warning in warnings:
                    print(warning)  # Each warning is a tuple (Level, Code, Message)

            temp_dir.cleanup()

        else:
            # For a small amount of data, I guess we can use the old way
            table_data = list(df0.itertuples(index=index))

            # https://mariadb-corporation.github.io/mariadb-connector-python/usage.html
            # "When using executemany(), there are a few restrictions:"

            # "1. Special values like None or column default value needs to
            #     be indicated by an indicator."
            MARIADB_NULL = mariadb.constants.INDICATOR.NULL

            def int_converter(v: Any) -> Any:
                if isinstance(v, np.int64):
                    return int(v)
                else:
                    return v

            try:
                table_data_typed: list[tuple[Any, ...]] = [
                    tuple(
                        [
                            MARIADB_NULL if pd.isna(value) else int_converter(value)
                            for value in sublist
                        ]
                    )
                    for sublist in table_data
                ]
                table_data = table_data_typed  # type: ignore[assignment]
            except ValueError as e:
                if get_env_bool("INTERACT_AFTER_ERROR"):
                    code.interact(
                        local=locals(), banner=f"DataFrameFast tupling error: {e}"
                    )
                else:
                    raise ValueError(e)

            # Optimization: just use `execute` if it's a single line of data
            if len(table_data) == 1:
                with con.cursor() as cursor:
                    try:
                        cursor.execute(cmd, table_data[0])
                    except Exception as e:
                        if get_env_bool("INTERACT_AFTER_ERROR"):
                            code.interact(
                                banner=f"DataFrameFast error: {e}", local=locals()
                            )
                        else:
                            raise Exception(e)
                return

            # 2. A workaround to https://jira.mariadb.org/browse/CONPY-254
            table_data_bad = [
                v for i, v in enumerate(table_data) if v[-1] == MARIADB_NULL
            ]
            table_data_good = [
                v for i, v in enumerate(table_data) if v[-1] != MARIADB_NULL
            ]

            with con.cursor() as cursor:
                # Try to use the bulk (faster) option for the "good" rows
                if len(table_data_good) > 0:
                    try:
                        cursor.executemany(cmd, table_data_good)
                    except mariadb.ProgrammingError as e:
                        print(
                            f"Warning: problem with executemany; probably because you are using a "
                            f"np.dtypes.Int64DType data type.  We'll just insert these one at a time. {e}"
                        )
                        for v in table_data_good:
                            cursor.execute(cmd, v)
                    except SystemError as e:
                        # 3. "All tuples must have the same types as in first tuple.
                        #    E.g. the parameter [(1),(1.0)] or [(1),(None)] are invalid."
                        print(
                            "Warning: due to a known bug in a mariadb connector, e.g. None in final field or something, "
                            f"`executemany` raises {e}.  Failing over to row-by-row `execute`"
                        )
                        for v in table_data_good:
                            cursor.execute(cmd, v)

                if len(table_data_bad) > 0:
                    for v in table_data_bad:
                        cursor.execute(cmd, v)

    def column_info(
        self, con: mariadb.Connection, table_name: str | None = None
    ) -> pd.DataFrame:
        """Returns the column information.
        Parameters:
            table_name: string.  If None, returns column info for ALL tables.
        """
        clauses: list[str] = [f"TABLE_SCHEMA = '{self.database}'"]
        if table_name is not None:
            clauses.append(f"TABLE_NAME = '{table_name}'")

        with con.cursor() as this_cursor:
            this_cursor.execute(
                f"SELECT * FROM INFORMATION_SCHEMA.COLUMNS "
                f"WHERE {'AND '.join(clauses)};"
            )
            records = this_cursor.fetchall()
            fields = this_cursor.description

        # Convert records to list of dicts
        records_dict = [
            {fields[i][0]: field_value for i, field_value in enumerate(v)}
            for v in records
        ]

        return pd.DataFrame(records_dict)


def read_sql_table(
    name: str,
    con: mariadb.Connection,
    query: str,
    *args: Any,
    **kwargs: Any,
) -> pd.DataFrame:
    """A drop-in replacement for pd.read_sql_table
    Note: this does not set an index.
    """
    assert query is not None

    # Use the main tonydbc to retrieve this with
    records, type_codes = con.get_data(
        query=query, no_tracking=True, return_type_codes=True
    )

    # TODO: convert all cols

    # KLUDGE: for now, just convert any BIT fields to bool
    bit_cols = {k: bool for k, v in type_codes.items() if v.upper() == "BIT"}

    if len(records) > 0:
        df = DataFrameFast.from_records(records)

        for k in bit_cols.keys():
            # Convert b'\x01' to True and b'\x00' to False
            df[k] = df[k].apply(
                lambda cell: bool(int.from_bytes(cell, byteorder="big"))
            )
    elif "CALL" not in query.upper():
        # We also have to return the columns names in case records is []
        """Use a trick to get the column names,
        even if the result set is 0 rows.  ChatGPT told me!  Wow.
        """
        with con.cursor() as cursor:
            cursor.execute(query)
            # Get the field names - we don't need the records for this case
            q0 = query.strip().rstrip(";")
            if " LIMIT " in q0.upper():
                cursor.execute(q0)
            else:
                cursor.execute(q0 + " LIMIT 0;")
            columns = [v[0] for v in cursor.description]

        df = DataFrameFast(columns=columns)
        df = df.astype(bit_cols)
    else:
        # We cannot get the columns if it was a stored procedure
        return DataFrameFast(columns=[])

    return df



# -*- coding: utf-8 -*-
########################################
## TEST HARNESS

import os
import pathlib

import pyperclip

from .tony_utils import prepare_scripts
from .tonydbc import TonyDBC


def create_test_database(
    test_db: str,
    source_db: str,
    schema_filepaths: list[str],
    tables_to_copy: dict[str, int],
    custom_commands: list[str],
    host: str,
    user: str,
    password: str,
    users_to_grant: list[str],
) -> None:
    """
    Drops the test_db, recreates all tables, copies the data from a source_db.

    Parameters:
        test_db:          The name of the database to DROP / CREATE
        source_db:        The name of the database to copy table rows from
        schema_filepaths: The SQL Data Definition commands to run.  Will be run in sequence.
        tables_to_copy:   The tables to copy from source_db, and the max # of rows to copy.
        custom_commands:  A list of commands to execute after tables are created
        host:             The IP or domain name to connect to.
        user:             The power user that can create databases, etc.
        password:
        users_to_grant:   The users that will be granted permission to access this test database.
    """
    for schema_filepath in schema_filepaths:
        if not os.path.isfile(schema_filepath):
            raise AssertionError(
                f"Schema file {schema_filepath} does not point to a valid file."
            )
        if not pathlib.Path(schema_filepath).suffix.lower() == ".sql":
            raise AssertionError(f"Schema file {schema_filepath} does not end in .sql.")

    program_to_run = ""

    with TonyDBC(host=host, user=user, password=password, database=source_db) as db:
        if test_db in db.production_databases:
            raise AssertionError(
                f"DANGER DANGER!  You set the test_db to be {test_db}, "
                f"one of our production databases ({db.production_databases})! "
                f"You were about to drop the production database! "
                f"Please talk to your manager for advice on what to do. "
                f"Do NOT bypass this error."
            )

        if test_db in db.databases:
            program_to_run += f"DROP DATABASE {test_db};\n"

    program_to_run += f"CREATE DATABASE {test_db};\n"
    for user_to_grant in users_to_grant:
        # No need to assign yourself permissions :D
        if not user_to_grant == user:
            program_to_run += (
                f"GRANT ALL ON {test_db}.* TO "
                f"'{user_to_grant}'@'%' WITH GRANT OPTION;\n"
            )
            # FYI: No need to "FLUSH PRIVILEGES;" if using the "GRANT" command
    program_to_run += f"USE {test_db};\n"

    # We must run the trigger script AFTER tables are copied since there might be
    # DB triggers which would fail during the INSERT commands above if the triggers
    # were already in place
    non_trigger_scripts = [f for f in schema_filepaths if str(f).find("trigger") == -1]
    trigger_scripts = [f for f in schema_filepaths if str(f).find("trigger") != -1]

    program_to_run += prepare_scripts(test_db, non_trigger_scripts)

    program_to_run += f"USE {test_db};\nSET FOREIGN_KEY_CHECKS = 0;\n"

    # Copy data from the source database
    # Prepend the source database to each entry, e.g. fling_db.wsc_event_log
    tables_to_copy = {
        ".".join([source_db, table]): max_rows
        for table, max_rows in tables_to_copy.items()
    }
    # Let's make it alphabetical for convenience during debugging
    # (note that e.g. "B" comes before "a", in lexicographical order, due to upper case being first)
    tables_to_copy = dict(sorted(tables_to_copy.items()))
    for table_string, max_rows in tables_to_copy.items():
        # Convert from 'src_db.table'
        assert "." in table_string
        cur_source_db, table = table_string.split(".")
        # print(f"COPY {table_string} -> {test_db}.{table}")
        cmd = (
            f"ALTER TABLE `{test_db}`.`{table}` DISABLE KEYS; "
            f"INSERT INTO `{test_db}`.`{table}` "
            f"SELECT * FROM `{cur_source_db}`.`{table}` LIMIT {max_rows}; "
            f"ALTER TABLE `{test_db}`.`{table}` ENABLE KEYS;\n"
        )
        program_to_run += cmd

        # Disabled due to pathetic bugginess and slowness
        # db.execute(
        #    cmd,
        #    before_retry_cmd=f"SET FOREIGN_KEY_CHECKS = 0; TRUNCATE {test_db}.{table};",
        # )
        # TODO: This command is causing 'mariadb.InterfaceError:
        # Lost connection to server during query'
        # The docs say this happens https://dev.mysql.com/doc/refman/8.0/en/gone-away.html
        # because of an INSERT statement that inserts a great many rows.
        # therefore we should find a way to insert in batches, or
        # maybe use a stored procedure or something - not sure if that would help
        # but just re-running the refresh seems to work since we only get the error
        # about 20% of the time.  It happens on wsc_status_log, the biggest table,
        # with 580k+ rows.
        pass

    program_to_run += "\n".join(custom_commands)

    program_to_run += prepare_scripts(test_db, trigger_scripts)

    program_to_run += "\nSET FOREIGN_KEY_CHECKS = 1;"

    pyperclip.copy(program_to_run)

    print(
        """
        SQL CODE has been COPIED TO YOUR CLIPBOARD:

        📂 ----> 📚 ---> 🗃️

        This code runs really slowly and sometimes aborts if we try to run it on
        the weak official mariadb Python connector.

        So instead, please:

        ** RUN it in HeidiSQL or your favourite Database Client.  Thank you! **
    """
    )



"""
Integration tests for TonyDBC using real MariaDB containers

This test file uses real MariaDB database containers (via testcontainers)
to test TonyDBC functionality with actual database interactions,
rather than mocking database calls.
"""

import os
import sys
import time
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Generator

import mariadb  # type: ignore
import pandas as pd
import pytest
from mariadb.constants.CLIENT import MULTI_STATEMENTS  # type: ignore

# Set required environment variables BEFORE importing tonydbc
os.environ.setdefault("USE_PRODUCTION_DATABASE", "False")
os.environ.setdefault("CHECK_ENVIRONMENT_INTEGRITY", "False")
os.environ.setdefault("INTERACT_AFTER_ERROR", "False")
os.environ.setdefault("DEFAULT_TIMEZONE", "UTC")
os.environ.setdefault("MYSQL_DATABASE", "test")
os.environ.setdefault("MYSQL_HOST", "localhost")
os.environ.setdefault("MYSQL_READWRITE_USER", "test")
os.environ.setdefault("MYSQL_READWRITE_PASSWORD", "test")
os.environ.setdefault("MYSQL_PRODUCTION_DATABASE", "test_prod")
os.environ.setdefault("MYSQL_TEST_DATABASE", "test_db")

# Add the src directory to the path so we can import tonydbc
sys.path.insert(0, str(Path(__file__).parent.parent / "src"))

from testcontainers.core.container import DockerContainer

import tonydbc


def _wait_db(
    host: str, port: int, user: str, pwd: str, db: str, timeout: int = 30
) -> None:
    """Wait for database to be ready"""
    start_time = time.time()
    while time.time() - start_time < timeout:
        try:
            conn = mariadb.connect(
                host=host,
                port=port,
                user=user,
                password=pwd,
                database=db,
                client_flag=MULTI_STATEMENTS,
            )
            conn.close()
            print(f"✅ Database ready at {host}:{port}")
            return
        except Exception as e:
            print(f"⏳ Waiting for database... ({e})")
            time.sleep(1)
    raise TimeoutError(f"Database not ready after {timeout} seconds")


@pytest.fixture(scope="session")
def mariadb_container() -> Generator[Any, None, None]:
    """Create a MariaDB container for testing"""

    user, pwd, db = "test", "test", "test"
    container = (
        DockerContainer("mariadb:10.6")
        .with_env("MYSQL_ROOT_PASSWORD", "root")
        .with_env("MYSQL_DATABASE", db)
        .with_env("MYSQL_USER", user)
        .with_env("MYSQL_PASSWORD", pwd)
        .with_exposed_ports(3306)
    )

    with container as c:
        host = c.get_container_host_ip()
        port = int(c.get_exposed_port(3306))
        _wait_db(host, port, user, pwd, db, timeout=120)

        yield SimpleNamespace(
            get_container_host_ip=lambda: host,
            get_exposed_port=lambda _p: port,
            username=user,
            password=pwd,
            dbname=db,
            container=c,
        )


@pytest.fixture(scope="session")
def tonydbc_instance(mariadb_container: Any) -> Any:
    """Create a TonyDBC instance connected to the test container"""
    container_host = mariadb_container.get_container_host_ip()
    container_port = mariadb_container.get_exposed_port(3306)

    # Override environment variables with actual container details
    container_env = {
        "MYSQL_HOST": container_host,
        "MYSQL_PORT": str(container_port),
        "MYSQL_READWRITE_USER": mariadb_container.username,
        "MYSQL_READWRITE_PASSWORD": mariadb_container.password,
        "MYSQL_DATABASE": mariadb_container.dbname,
        "MYSQL_TEST_DATABASE": mariadb_container.dbname,
        "DEFAULT_TIMEZONE": "UTC",
    }

    # Apply the container environment
    original_env = {}
    for key, value in container_env.items():
        original_env[key] = os.environ.get(key)
        os.environ[key] = value

    try:
        # Create TonyDBC instance
        db = tonydbc.TonyDBC(
            host=container_host,
            port=container_port,
            user=mariadb_container.username,
            password=mariadb_container.password,
            database=mariadb_container.dbname,
        )
        yield db
    finally:
        # Restore original environment
        for key, value in original_env.items():
            if value is None:
                os.environ.pop(key, None)
            else:
                os.environ[key] = value


@pytest.fixture
def setup_tables(tonydbc_instance):
    """Set up test tables in the database"""
    with tonydbc_instance as db:
        # Create a table with a primary key
        db.execute("""
            CREATE TABLE IF NOT EXISTS test_table (
                id INT AUTO_INCREMENT PRIMARY KEY,
                name VARCHAR(100),
                email VARCHAR(100)
            )
        """)

        # Create a table with NO primary key
        db.execute("""
            CREATE TABLE IF NOT EXISTS no_pk_table (
                col1 VARCHAR(100),
                col2 INT
            )
        """)

        # Create a users table
        db.execute("""
            CREATE TABLE IF NOT EXISTS users (
                user_id INT AUTO_INCREMENT PRIMARY KEY,
                username VARCHAR(50),
                email VARCHAR(100)
            )
        """)

        yield db

        # Cleanup
        db.execute("DROP TABLE IF EXISTS test_table")
        db.execute("DROP TABLE IF EXISTS no_pk_table")
        db.execute("DROP TABLE IF EXISTS users")


class TestTonyDBCIntegration:
    """Integration tests using real MariaDB containers"""

    def test_get_primary_key_with_real_table(self, setup_tables):
        """Test get_primary_key with real database table"""
        db = setup_tables

        # Test table with primary key
        result = db.get_primary_key("test_table")
        assert result == "id"

        # Test table with primary key (users)
        result = db.get_primary_key("users")
        assert result == "user_id"

    def test_get_primary_key_with_default_real_table(self, setup_tables):
        """Test get_primary_key returns default when table has no primary key"""
        db = setup_tables

        # Test table with NO primary key - should return default
        result = db.get_primary_key("no_pk_table", default="pk_id")
        assert result == "pk_id"

    def test_insert_and_query_real_data(self, setup_tables):
        """Test inserting and querying real data"""
        db = setup_tables

        # Insert some test data
        db.execute(
            "INSERT INTO test_table (name, email) VALUES ('John', 'john@test.com')"
        )
        db.execute(
            "INSERT INTO test_table (name, email) VALUES ('Jane', 'jane@test.com')"
        )

        # Query the data back
        result = db.get_data("SELECT * FROM test_table ORDER BY id")

        assert len(result) == 2
        assert result[0]["name"] == "John"
        assert result[0]["email"] == "john@test.com"
        assert result[1]["name"] == "Jane"
        assert result[1]["email"] == "jane@test.com"

    def test_dataframe_operations_real_db(self, setup_tables):
        """Test DataFrame operations with real database"""
        db = setup_tables

        # Create test DataFrame
        df = pd.DataFrame(
            {
                "name": ["Alice", "Bob", "Charlie"],
                "email": ["alice@test.com", "bob@test.com", "charlie@test.com"],
            }
        )

        # Write DataFrame to database
        db.write_dataframe(df, "test_table", if_exists="append", index=False)

        # Read data back
        result_df = db.query_table("test_table", "SELECT * FROM test_table")

        assert len(result_df) == 3
        assert "Alice" in result_df["name"].values
        assert "Bob" in result_df["name"].values
        assert "Charlie" in result_df["name"].values

    def test_append_to_table_real_db(self, setup_tables):
        """Append rows and verify persisted results"""
        db = setup_tables

        # Start from a clean slate
        db.execute("DELETE FROM test_table;")

        df = pd.DataFrame(
            {
                "name": ["Alpha", "Beta"],
                "email": ["alpha@test.com", "beta@test.com"],
            }
        )

        db.append_to_table("test_table", df)

        rows = db.get_data("SELECT name, email FROM test_table ORDER BY name;")
        assert [(r["name"], r["email"]) for r in rows] == [
            ("Alpha", "alpha@test.com"),
            ("Beta", "beta@test.com"),
        ]

    def test_read_dataframe_from_table_real_db(self, setup_tables):
        """Read DataFrame using explicit query and verify shape"""
        db = setup_tables

        # Ensure at least one row exists
        db.execute(
            "INSERT INTO test_table (name, email) VALUES ('Zed', 'zed@test.com')"
        )

        df = db.read_dataframe_from_table(
            table_name="test_table", query="SELECT id, name, email FROM test_table;"
        )
        assert isinstance(df, pd.DataFrame)
        # Primary key is intentionally set as index when named 'id'
        assert df.index.name == "id"
        assert set(["name", "email"]).issubset(set(df.columns))

    def test_update_table_real_db(self, setup_tables):
        """Update rows via DataFrame indexed by PK and verify changes"""
        db = setup_tables

        # Seed two rows
        db.execute("DELETE FROM test_table;")
        db.execute(
            "INSERT INTO test_table (name, email) VALUES ('U1', 'u1@test.com'), ('U2', 'u2@test.com')"
        )
        rows = db.get_data("SELECT id, name, email FROM test_table ORDER BY id;")
        ids = [r["id"] for r in rows]

        # Build update dataframe indexed by PK
        upd = pd.DataFrame(
            {"name": ["U1x", "U2y"], "email": ["x@test.com", "y@test.com"]}
        )
        upd.index = ids
        upd.index.name = "id"

        db.update_table("test_table", upd)

        rows2 = db.get_data("SELECT id, name, email FROM test_table ORDER BY id;")
        assert [(r["name"], r["email"]) for r in rows2] == [
            ("U1x", "x@test.com"),
            ("U2y", "y@test.com"),
        ]

    def test_query_table_real_db(self, setup_tables):
        """Query with explicit SQL and verify returned DataFrame"""
        db = setup_tables
        df = db.query_table("test_table", "SELECT id, name FROM test_table LIMIT 1;")
        # 'id' is used as index by design
        assert df.index.name == "id"
        assert "name" in df.columns

    @pytest.mark.skip(reason="Temporarily skipped due to 'replace' handling semantics")
    def test_write_dataframe_integration_results(self, setup_tables):
        """End-to-end: write_dataframe persists rows and values correctly"""
        db = setup_tables

        # Fresh table for this test
        db.execute("DROP TABLE IF EXISTS t;")
        db.execute("CREATE TABLE t (id INT PRIMARY KEY, name VARCHAR(32));")

        df = pd.DataFrame({"id": [1, 2], "name": ["A", "B"]})
        db.write_dataframe(df, "t", if_exists="replace", index=False)

        rows = db.get_data("SELECT id, name FROM t ORDER BY id;")
        assert [(r["id"], r["name"]) for r in rows] == [(1, "A"), (2, "B")]

    def test_column_datatypes_real_table(self, setup_tables):
        """Test getting column datatypes from real table"""
        db = setup_tables

        # Get column datatypes for test_table
        result = db.get_column_datatypes("test_table")

        assert "id" in result
        assert "name" in result
        assert "email" in result
        # The exact types may vary, but they should be present
        assert result["id"] in [int, "int"]
        assert result["name"] in [str, "str"]
        assert result["email"] in [str, "str"]

    def test_temp_id_table_real_db(self, setup_tables):
        """Test creating temporary ID table with real database"""
        db = setup_tables

        id_list = [1, 2, 3, 4, 5]

        # Use temp_id_table context manager
        with db.temp_id_table(id_list) as temp_table_name:
            # Verify the temporary table exists and has data
            result = db.get_data(f"SELECT * FROM {temp_table_name} ORDER BY id")

            assert len(result) == 5
            assert [row["id"] for row in result] == [1, 2, 3, 4, 5]

            # Verify table name format
            assert temp_table_name.startswith("temp_loc_ids_")

        # After context manager, table should be dropped
        # This should raise an error since table no longer exists
        with pytest.raises(Exception):
            db.get_data(f"SELECT * FROM {temp_table_name}")



"""
tonydbc

    TonyDBC: A context manager for mariadb data connections.

    MQTTClient: A context manager for an MQTT client

    create_test_database: function to create a complete test database

Note: you should define the following environment variables before using this library, e.g.

    CHECK_ENVIRONMENT_INTEGRITY = True
    DOT_ENVS                    = ["..\\..\\.env", ".env"]
    USE_PRODUCTION_DATABASE     = True
    PRODUCTION_DATABASES        = ["master_db", "master2_db", "etc"]
    MYSQL_TEST_DATABASE         = test_db
    MYSQL_PRODUCTION_DATABASE   = master_db
    INTERACT_AFTER_ERROR        = False
    AUDIT_PATH                  = database
    # Full list is at pytz.all_timezones
    DEFAULT_TIMEZONE            = Asia/Singapore
    MEDIA_BASE_PATH_PRODUCTION  = C:\\

e.g.
    import tonydbc
    tonydbc.load_dotenvs()

"""

__version__ = "1.3.3"

# Public API - explicitly declare what should be available when importing this package
__all__ = [
    "load_dotenv",
    "get_env_bool",
    "get_env_list",
    "load_dotenvs",
    "set_MYSQL_DATABASE",
    "get_current_time",
    "deserialize_table",
    "list_to_SQL",
    "list_to_SQL2",
    "TonyDBC",
    "MQTTClient",
    "DataFrameFast",
    "create_test_database",
]

# Include vanialla dotenv.load_dotenv to be comprehensive, but it's not really needed
# since the version people should use is load_dotenvs
from dotenv import load_dotenv

from .create_test_database import create_test_database
from .dataframe_fast import DataFrameFast
from .env_utils import get_env_bool, get_env_list, load_dotenvs
from .mqtt_client import MQTTClient
from .tony_utils import (
    deserialize_table,
    get_current_time,
    list_to_SQL,
    list_to_SQL2,
    set_MYSQL_DATABASE,
)
from .tonydbc import TonyDBC
